/*
  $Id: cmd.c 714 2006-03-07 10:21:18Z jens $

  Copyright (C) 2001-2006 Jens Thoms Toerring
 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
  To contact the author send email to: jt@toerring.de
*/


#include "gpib_module.h"
#include "sys.h"
#include "board.h"
#include "atgpib.h"


static unsigned char nff;


static void board_wait_cmd( Board_t * board );


/*-----------------------------------------------------------------*
 * Function writes a command to the bus. It must only be called if
 * board is ACIC, i.e. is controller and the ATN line is raised!
 *-----------------------------------------------------------------*/

int board_cmd( Board_t * board )
{
	unsigned int count = 0;
	int ccount;
	bool stop_timer = 0;


	PDEBUGin( );

	err = done = 0;

	/* Clear ISR0, ISR1 and ISR2 register and FIFO */

	board_in( reg_addr.ISR0 );
	board_in( reg_addr.ISR1 );
	board_in( reg_addr.ISR2 );
	board_out( reg_addr.CMDR, RESET_FIFO );

	/* Set up board to send a command in 8 bit transfer mode and to stop
	   transfer on ERR condition (i.e. no listener) */

	regs.cfg |= CFG_CMD | CFG_TLCHLTE;
	regs.cfg &= ~ ( CFG_IN | CFG_16_8N | CFG_CCEN | CFG_A_BN );
	board_out( reg_addr.CFG, regs.cfg );

	/* Make sure no DMA is used and clear IMR0 and IMR2 */

	if ( regs.accwr & ACCWR_DMAEN ) {
		regs.accwr &= ~ ACCWR_DMAEN;
		board_out( reg_addr.ACCWR, regs.accwr );
	}

	if ( regs.imr0 & 0x7F ) {
		regs.imr0 = 0x80;
		board_out( reg_addr.IMR0, regs.imr0 );
	}

	if ( regs.imr2 ) {
		regs.imr2 = 0;
		board_out( reg_addr.IMR2, regs.imr2 );
	}

	/* Write 2's complement of length of command to be send into the
	   counter registers */

	ccount = - board->transf.cnt;
	board_out( reg_addr.CNT0,
		   ( unsigned char ) ( ccount & 0xFF ) );
	board_out( reg_addr.CNT1,
		   ( unsigned char ) ( ( ccount >>  8 ) & 0xFF ) );
	board_out( reg_addr.CNT2,
		   ( unsigned char ) ( ( ccount >> 16 ) & 0xFF ) );
	board_out( reg_addr.CNT3,
		   ( unsigned char ) ( ( ccount >> 24 ) & 0xFF ) );

	/* If no timer is running start it now */

	if ( ! ( board->status & TIMER_ON ) ) {
		gpib_start_timer( );
		stop_timer = SET;
	}

	/* Enable interrupts (if they're to be used) to make the board flag
	   errors, the DONE condition and interrupts from IMR0, IMR1 or IMR2.
	   Then set up the transfer semaphore and switch on interrupts */

	if ( board->status & USE_IRQ ) {
		semaphore_init( &board->transf.sem, 0 );

		regs.imr1 |= IMR1_ERR_IE;
		board_out( reg_addr.IMR1, regs.imr1 );

		regs.imr3 |= IMR3_TLCINT_IE | IMR3_DONE_IE;
		board_out( reg_addr.IMR3, regs.imr3 );

		regs.intr |= INTR_INTEN;
		board_out( reg_addr.INTR, regs.intr );
	}

	/* Send the GO command */

	board_out( reg_addr.CMDR, GO );

	/* Write out all bytes into the FIFO using 8-bit PIO mode */

	PDEBUG( "Starting PIO write loop\n" );

	while ( count < board->transf.cnt ) {
		if ( ! ( nff = board_in( reg_addr.ISR3 ) & ISR3_NFF ) )
			board_wait_cmd( board );

		if ( done || err || board->timeout )
			goto cmd_done;

		board_out( reg_addr.FIFOB, * ( board->transf.buf + count++ ) );
	}

	/* Wait for transfer to finish */

	while ( ! ( done || err || board->timeout ) )
		board_wait_done( board );

 cmd_done:

	/* Send STOP command */

	board_out( reg_addr.CMDR, STOP );

	/* Stop timer if necessary */

	if ( stop_timer )
		gpib_stop_timer( );

	/* Disable interrupts (if they were used) */

	if ( board->status & USE_IRQ ) {
		regs.intr &= ~ INTR_INTEN;
		board_out( reg_addr.INTR, regs.intr );

		regs.imr1 &= ~ IMR1_ERR_IE;
		board_out( reg_addr.IMR1, regs.imr1 );

		regs.imr3 &= ~ ( IMR3_NFF_IE | IMR3_TLCINT_IE | IMR3_DONE_IE );
		board_out( reg_addr.IMR3, regs.imr3 );
	}

	/* Reset CFG to normal state */

	regs.cfg &= ~ ( CFG_CMD | CFG_TLCHLTE );
	board_out( reg_addr.CFG, regs.cfg );

	/* Calculate how many bytes have been written */

	count = - (   ( board_in( reg_addr.CNT0 ) & 0xFF )	      |
		    ( ( board_in( reg_addr.CNT1 ) & 0xFF ) <<  8 ) |
		    ( ( board_in( reg_addr.CNT2 ) & 0xFF ) << 16 ) |
		    ( ( board_in( reg_addr.CNT3 ) & 0xFF ) << 24 ) );
	gpib_count = board->transf.cnt - count;

	PDEBUG( "Wrote %ld of %d bytes\n", gpib_count, board->transf.cnt );

	/* Set the result flags */

	if ( err ) {
		gpib_status |= GPIB_ERR;
		if ( board->timeout )
			gpib_status |= GPIB_TIMO;
		gpib_error   = GPIB_ENOL;
		PDEBUG( "No listener in cmd\n" );
	} else if ( board->timeout ) {
		gpib_error   = GPIB_EABO;
		gpib_status |= GPIB_ERR | GPIB_TIMO;
		PDEBUG( "Timeout in cmd\n" );
	} else if ( board->transf.is_last && count == 0 )
		gpib_status |= GPIB_CMPL;

	PDEBUGout( );
	return gpib_status;
}


/*------------------------------------------------------------------*
 * When using interrupts do a down() on the transfer semaphore - it
 * will be up()'ed either by the interrupt handler or on timeout.
 * Without interrupts just poll register ISR1 and ISR3 until either
 * the FIFO isn't full anymore, the DONE or ERR condition is met or
 * a timeout happens.
 *------------------------------------------------------------------*/

static void board_wait_cmd( Board_t * board )
{
	unsigned char isr3;


	PDEBUGin( );

	if ( board->status & USE_IRQ ) {

		/* Enable interrupt on "not full FIFO" condition, i.e. if
		   further data can be written to the FIFO */

		regs.imr3 |= IMR3_NFF_IE;
		board_out( reg_addr.IMR3, regs.imr3 );

		/* Wait until semaphore gets up()'ed by the interrupt or
		   timer handler */

		if ( ! board->timeout )
			down( &board->transf.sem );

		/* Disable interrupt on "not full FIFO" condition */

		regs.imr3 &= ~ IMR3_NFF_IE;
		board_out( reg_addr.IMR3, regs.imr3 );

		/* Get state and error bits */

		isr3 = board_in( reg_addr.ISR3 );
		nff  = isr3 & ISR3_NFF;
		done = isr3 & ISR3_DONE;
		err  = board_in( reg_addr.ISR1 ) & ISR1_ERR;
	} else
		/* Poll the relevant registers until a byte can be written
		   or an error or timeout happened */

		do {
			isr3 = board_in( reg_addr.ISR3 );
			nff  = isr3 & ISR3_NFF;
			done = isr3 & ISR3_DONE;
			err  = board_in( reg_addr.ISR1 ) & ISR1_ERR;
		} while ( ! ( nff || done || err || board->timeout ) );

	PDEBUGout( );
}


/*
 * Local variables:
 * c-basic-offset: 8
 * c-indent-level: 8
 * c-brace-imaginary-offset: 0
 * c-brace-offset: 0
 * c-argdecl-indent: 4
 * c-label-offset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: t
 * tab-width: 8
 * End:
 */
