/*
  $Id: dma.c 714 2006-03-07 10:21:18Z jens $

  Copyright (C) 2001-2006 Jens Thoms Toerring
 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
  To contact the author send email to: jt@toerring.de
*/


#include "gpib_module.h"
#include "sys.h"
#include "board.h"
#include "atgpib.h"



/*-------------------------------------------------------------------------*
 * Function tries to request an interrupt and a DMA channel for the board.
 * If this fails no DMA can be used for transfers. The function returns an
 * error if either the IRQ or the DMA requested is invald.
 *-------------------------------------------------------------------------*/

bool board_irq_dma_init( Board_t * board )
{
	/* Do nothing if the board isn't set up to use interrupts - in this
	   case DMA also can't be used */

	if ( ! board->use_irq ) {
		board->use_dma = UNSET;
		board->status &= ~ ( USE_IRQ | USE_DMA );
		return OK;
	}

	/* Check that interrupt is reasonable (eventhough there's a
	   jumper on the board for IRQ 9 it's not connected !) */

	if ( board->irq < 3 || board->irq == 8 || board->irq == 9 ||
	     board->irq == 13 || board->irq > 15 ) {
		board->use_irq = UNSET;
		board->use_dma = UNSET;
		board->status &= ~ ( USE_IRQ | USE_DMA );
		printk( KERN_ERR "%s: Invalid IRQ %u requested.\n",
			board->name, board->irq );
		return FAIL;
	}

	/* Try to request the interrupt, if this fails we also can't use DMA */

	if ( request_irq( board->irq, gpib_interrupt_handler,
#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 6, 21 )
			  IRQF_SHARED,
#else
#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 4, 0 )
			  SA_SHIRQ,
#else
			  SA_INTERRUPT,
#endif
#endif
			  board->name, board ) ) {
		board->use_irq = UNSET;
		board->use_dma = UNSET;
		board->status &= ~ ( USE_IRQ | USE_DMA );
		printk( KERN_WARNING "%s: Can't get assigned IRQ %u, won't "
			"use interrupts or DMA.\n", board->name, board->irq );
		return OK;
	}

	board->status |= USE_IRQ;

	/* We're done if DMA is not to be used */

	if ( ! board->use_dma ) {
		board->status &= ~ USE_DMA;
		return OK;
	}

	/* Check that DMA channel is reasonable (the board can only use the
	   16-bit DMA channels between 5 and 7) */

	if ( board->use_dma && ( board->dma < 5 || board->dma > 7 ) ) {
		board->use_dma = UNSET;
		board->status &= ~ USE_DMA;
		printk( KERN_ERR "%s: Invalid DMA channel %u requested.\n",
			board->name, board->dma );
		return FAIL;
	}

	/* Now request the  DMA channel */

	if ( request_dma( board->dma, board->name ) ) {
		board->use_dma = UNSET;
		board->status &= ~ USE_DMA;
		printk( KERN_WARNING "%s: Can't get assigned DMA channel %u, "
			"won't use DMA.\n", board->name, board->dma );
	} else
		board->status |= USE_DMA;

	return OK;
}


/*-------------------------------------------------------------------------*
 * Function frees the DMA channel and IRQ level if they had been requested
 *-------------------------------------------------------------------------*/

void board_irq_dma_cleanup( Board_t * board )
{
	if ( board->use_dma )
		free_dma( board->dma );
	if ( board->use_irq )
		free_irq( board->irq, board );
}


/*
 * Local variables:
 * c-basic-offset: 8
 * c-indent-level: 8
 * c-brace-imaginary-offset: 0
 * c-brace-offset: 0
 * c-argdecl-indent: 4
 * c-label-offset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: t
 * tab-width: 8
 * End:
 */
