/*
  $Id: init.c 731 2006-05-29 10:02:36Z jens $

  Copyright (C) 2001-2006 Jens Thoms Toerring
 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
  To contact the author send email to: jt@toerring.de
*/


#include "gpib_module.h"
#include "autoconf.h"
#include "sys.h"
#include "board.h"
#include "atgpib.h"


unsigned char err;
unsigned char done;
TNT4882_Register_Adresses_t reg_addr;
TNT4882_Register_Contents_t regs;


static bool board_request_region( Board_t * board );
static void board_release_region( Board_t * board );
static bool board_init_chip( Board_t * board );


/*---------------------------------------------------------*
 * Does everything to initialize the board: The I/O region
 * for the board gets locked and the chip initialized.
 *---------------------------------------------------------*/

bool board_init( Board_t * board )
{
	/* Store the board type */

	board->type = ATGPIB;

	/* Get a lock on the memory locations used by the board and set up the
	   register pointers to these locations */ 

	if ( ! board_request_region( board ) )
		return FAIL;

	/* Initialize the chip */

	if ( ! board_init_chip( board ) )
	{
		board_release_region( board );
		return FAIL;
	}

	/* Set EOS character and eosmode (the values to be set have both
	   already been set up) */

	board_set_eos( board );
	board_set_eosmode( board, 0 );
	board->status &= ~ HOLDOFF;

	return OK;
}


/*-----------------------------------------------------------------------*
 * Called when the module is becoming unloaded: the board goes to Idle
 * Controller State and the I/O region locked for the board is released.
 *-----------------------------------------------------------------------*/

void board_cleanup( Board_t * board )
{
	/* Make board to Idle Controller state */

	board_release_control( board );

	/* Release I/O region */

	board_release_region( board );
}


/*----------------------------------------------*
 * Requests the I/O region needed for the board
 *----------------------------------------------*/

static bool board_request_region( Board_t * board )
{
	unsigned char *base;


	/* The board needs an I/O address range of 32 byte */

	board->io_size = 0x20;

	/* Check that the boards base address is valid */

	for ( base  = ( unsigned char * ) 0x0100;
	      base <= ( unsigned char * ) 0x03e0;
	      base += board->io_size )
		if ( base == ( unsigned char * ) board->base )
			break;

	if ( base > ( unsigned char * ) 0x0400 ) {
		printk( KERN_ERR "%s: Invalid base address 0x%lx for "
			"atgpib board.\n", board->name, board->base );
		return FAIL;
	}

	/* Check that the 32 byte at the boards address are not already
	   in use by someone else, then lock them */

#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 6, 0 )
	if ( request_region( board->base, board->io_size,
			     board->name ) == NULL ) {
		printk( KERN_ERR "%s: Can't get region at boards base "
			"address 0x%lu\n", board->name, board->base );
		return FAIL;
	}
#else
	if ( check_region( board->base, board->io_size ) < 0 ) {
		printk( KERN_ERR "%s: Can't get region at boards base "
			"address 0x%lu\n", board->name, board->base );
		return FAIL;
	}

	request_region( base, board->io_size, board->name );
#endif

	/* Set up the pointers to the memory locations associated with the
	   chips registers */

	reg_addr.DIR	       = base++;
	reg_addr.dont_touch_1  = base++;
	reg_addr.ISR1	       = base++;
	reg_addr.dont_touch_3  = base++;
	reg_addr.ISR2	       = base++;
	reg_addr.dont_touch_5  = base++;
	reg_addr.SPSR	       = base++;
	reg_addr.dont_touch_7  = base++;
	reg_addr.ADSR	       = base++;
	reg_addr.CNT2	       = base++;
	reg_addr.CPTR	       = base++;
	reg_addr.CNT3	       = base++;
	reg_addr.ADR0	       = base++;
	reg_addr.dont_touch_D  = base++;
	reg_addr.ADR1	       = base++;
	reg_addr.dont_touch_F  = base++;
	reg_addr.STS1	       = base++;
	reg_addr.DSR	       = base++;
	reg_addr.IMR3	       = base++;
	reg_addr.dont_touch_13 = base++;
	reg_addr.CNT0	       = base++;
	reg_addr.dont_touch_15 = base++;
	reg_addr.CNT1	       = base++;
	reg_addr.CSR	       = base++;
	reg_addr.FIFOB	       = base++;
	reg_addr.FIFOA	       = base++;
	reg_addr.ISR3	       = base++;
	reg_addr.SASR	       = base++;
	reg_addr.STS2	       = base++;
	reg_addr.ISR0	       = base++;
	reg_addr.TIMER	       = base++;
	reg_addr.BSR	       = base++;

	return OK;
}


/*----------------------------------------------*
 * Releases the I/O region locked for the board
 *----------------------------------------------*/

static void board_release_region( Board_t * board )
{
	release_region( ( PORT ) board->base, board->io_size );
}


/*-------------------------------------------------*
 * Initialization of board - does a reset and sets
 * all the needed parameter to the chips registers
 *-------------------------------------------------*/

static bool board_init_chip( Board_t * board )
{
	unsigned char signature;


	/* Read the signature register. If the documentation is correct the
	   upper four bits should be a number between 3 and 7 for all
	   versions of the TNT4882. The two lowest bits should always be 0.
	   By checking that bits 7, 1 and 0 are zero we can at least guess if
	   there's a chip at all (but which doesn't guarantee that the chip
	   exists, further checks are down later). Next we test the value of
	   the upper four bits to find out if this is a supported version of
	   the chip. */

	signature = ( board_in( reg_addr.CSR ) & CSR_SIGNATURE );
	if ( signature & 0x83 ) {
		printk( KERN_ERR "%s: Failed to initialize chip, not a "
			"TNT4882C or board is missing.\n", board->name );
		return FAIL;
	}

	signature >>=  4;
	PDEBUG( "Chip signature is %x\n", signature );

	if ( signature < 3 ) {
		printk( KERN_ERR "%s: Wrong chip signature, not a "
			"TNT4882C.\n", board->name );
		return FAIL;
	}

	/* Do software reset, release control and then do a chip reset */

	board_out( reg_addr.CMDR, SOFT_RESET );

	regs.cfg   = 0;
	regs.hssel = 0;
	regs.imr3  = 0;

	board_out( reg_addr.CMDR, CLRSC );

	board_out( reg_addr.AUXMR, AUX_ch_rst );
	udelay( 1 );

	regs.spmr  = 0;
	regs.timer &= ~ 0x3;
	regs.auxra = regs.auxrb = regs.auxre = regs.auxrf
		   = regs.auxrg = regs.auxri = regs.auxrj = 0;
	regs.ppr  = 0;
	regs.misc = 0;
	regs.hier = 0;
	regs.pt1  = 0;

	/* Switch to one-chip mode */

	regs.hssel = HSSEL_ONEC;
	board_out( reg_addr.HSSEL, regs.hssel );

	/* No use of DMA yet (but should be cleared anyway due to reset) */

	regs.accwr &= ~ ACCWR_DMAEN;
	board_out( reg_addr.ACCWR, regs.accwr );

	/* Before really setting the address of the board we use the address
	   registers to test if this is really a TNT4882 - if the address we
	   write to the address register isn't identical to the one we get
	   back this is either the wrong chip or there's no chip at all. */

	board_out( reg_addr.ADMR, 0x30 | ADMR_ADM0 );

	board_out( reg_addr.ADR, 0x60 | 0x15 );
	if ( ( board_in( reg_addr.ADR0 ) & 0x1F ) != 0x15 ) {
		printk( KERN_ERR "%s: Failed to initialize chip, not a "
			"TNT4882C or board is missing.\n", board->name );
		return FAIL;
	}

	board_out( reg_addr.ADR, 0x60 | 0x0A );
	if ( ( board_in( reg_addr.ADR0 ) & 0x1F ) != 0x0A ) {
		printk( KERN_ERR "%s: Failed to initialize chip, not a "
			"TNT4882C or board is missing.\n", board->name );
		return FAIL;
	}

	board_out( reg_addr.ADR, 0xE0 | 0x15 );
	if ( ( board_in( reg_addr.ADR1 ) & 0x1F ) != 0x15 ) {
		printk( KERN_ERR "%s: Failed to initialize chip, not a "
			"TNT4882C or board is missing.\n", board->name );
		return FAIL;
	}

	board_out( reg_addr.ADR, 0xE0 | 0x0A );
	if ( ( board_in( reg_addr.ADR1 ) & 0x1F ) != 0x0A ) {
		printk( KERN_ERR "%s: Failed to initialize chip, not a "
			"TNT4882C or board is missing.\n", board->name );
		return FAIL;
	}

	/* Set normal dual addressing mode, then set the primary address and,
	   if given (i.e. it's a secondary address), the secondary address,
	   otherwise disable use of secondary address */

	board_out( reg_addr.ADMR, 0x30 | ADMR_ADM0 );
	regs.admr |= 0x30 | ADMR_ADM0;
	board_out( reg_addr.ADR, board->pad & 0x1F );

	if ( board->sad >= 0x60 && board->sad <= 0x7E )
		board_out( reg_addr.ADR, ADR_ARS | ( board->sad & 0x1F ) );
	else
		board_out( reg_addr.ADR, ADR_ARS | ADR_DT | ADR_DL );

	/* Disable reactions to parallel polls (for now we assume we're the
	   only controller connected to the bus) */

	regs.ppr |= PPR_U;
	board_out( reg_addr.AUXMR, PPR | regs.ppr );
	udelay( 1 );

	regs.auxri |= AUXRI_PP2;
	board_out( reg_addr.AUXMR, AUXRI | regs.auxri );
	udelay( 1 );

	/* Don't allow interrupts yet and clear interrupt enable bits in IMR0,
	   IMR1 and IMR2 (IMR3 is already cleared by the software reset */

	regs.intr &= ~ INTR_INTEN;
	board_out( reg_addr.INTR, regs.intr );

	regs.imr0 = 0x80;
	board_out( reg_addr.IMR0, regs.imr0 );
	regs.imr1 = 0;
	board_out( reg_addr.IMR1, regs.imr1 );
	regs.imr2 = 0;
	board_out( reg_addr.IMR2, regs.imr2 );

	/* Don't use HS488 mode */

	regs.misc = 0;
	board_out( reg_addr.HIER, regs.misc );

	/* Let's be careful and use the slowest deglitching for the recognition
	   of DAV signals */

	regs.hier |= HIER_DGA | HIER_DGB;
	board_out( reg_addr.HIER, regs.hier );

	/* Also use slower T1 timing of 1100 ns (T1 is the delay between
	   setting the data lines and sending DAV) */

	regs.keyreg &= ~ KEYREG_MSTD;
	board_out( reg_addr.KEYREG, regs.keyreg );
	regs.auxri |= AUXRI_USTD;
	board_out( reg_addr.AUXMR, AUXRI | regs.auxri );
	udelay( 1 );

	/* Set timeout for DMA operation - needs to be set to the two's
	   complement of the time in units of 100 ns, so 0x06 stands for
	   a timeout value of 25 us. */

	regs.timer = 0x06;
	board_out( reg_addr.TIMER, regs.timer );

	regs.cfg |= CFG_TMOE | CFG_TIM_BYTN;
	board_out( reg_addr.CFG, regs.cfg );

	/* Finally release pon state that was entered due to the chip reset */

	board_out( reg_addr.AUXMR, AUX_pon );
	udelay( 1 );

	return OK;
}


/*-------------------------------------------------------------------*
 * Makes the board release its control over the GPIB bus by going to
 * the Idle Controller state. If it's alredy in this state calling
 * this function does nothing.
 *-------------------------------------------------------------------*/

int board_release_control( Board_t * board )
{
	unsigned char state;

	PDEBUGin( );


	state =	 board_get_state( board );

	if ( board->status & REN )
		board_ren( board, 0 );

	if ( state & GPIB_ACIC )
		board_gts( board );

	if ( state & GPIB_CIC )
		board_out( reg_addr.CMDR, CLRSC );

	gpib_status = state & ~ ( GPIB_ATN | GPIB_CIC );

	PDEBUGout( );
	return gpib_status;
}


/*
 * Local variables:
 * c-basic-offset: 8
 * c-indent-level: 8
 * c-brace-imaginary-offset: 0
 * c-brace-offset: 0
 * c-argdecl-indent: 4
 * c-label-ofset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: t
 * tab-width: 8
 * End:
 */
