/*
  $Id: read.c 732 2006-05-30 20:13:39Z jens $

  Copyright (C) 2001-2006 Jens Thoms Toerring
 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
  To contact the author send email to: jt@toerring.de
*/


#include "gpib_module.h"
#include "sys.h"
#include "board.h"
#include "atgpib.h"


static unsigned char end;
static unsigned char nef;


static int board_dma_read( Board_t * board );
static int board_pio_read( Board_t * board );


/*---------------------------------------------------------------------------*
 * If interrupts can be used a down() is done on the transfer semaphore -
 * if the interrupt has already come in this does nothing because then up()
 * has already been called on the semaphore by the interrupt handler, other-
 * wise the process will sleep until the interrupt handler is triggered. If
 * no interrupts can be used the ISR3 and ISR1 registers are polled until
 * a new byte can be read or a timeout or error occured.
 *---------------------------------------------------------------------------*/

inline static void board_wait_in( Board_t * board )
{
	unsigned char isr3;


	PDEBUGin( );

	if ( board->status & USE_IRQ ) {

		/* Enable interrupt on "not empty FIFO" condition, i.e if
		   there are data to read */

		regs.imr3 |= IMR3_NEF_IE;
		board_out( reg_addr.IMR3, regs.imr3 );		

		/* Wait until semaphore gets up()'ed by the interrupt or
		   timer handler */

		if ( ! board->timeout )
			down( &board->transf.sem );

		/* Disable interrupt on "not empty FIFO" condition */

		regs.imr3 &= ~ IMR3_NEF_IE;
		board_out( reg_addr.IMR3, regs.imr3 );

		/* Get state and error bits */

		isr3 = board_in( reg_addr.ISR3 );
		nef  = isr3 & ISR3_NEF;
		done = isr3 & ISR3_DONE;
		end  = board_in( reg_addr.ISR1 ) & ISR1_END_RX;
	} else
		/* Poll the relevant registers until a byte can be read
		   or an error or timeout happened */

		do {
			isr3 = board_in( reg_addr.ISR3 );
			nef  = isr3 & ISR3_NEF;
			done = isr3 & ISR3_DONE;
			end  = board_in( reg_addr.ISR1 ) & ISR1_END_RX;
		} while ( ! ( nef || done || end || board->timeout ) );

	PDEBUGout( );
}


/*-------------------------------------------------------------------------*
 * Function decides if the transfer is to be done using DMA or PIO.
 * Unfortunately, it is not possible to resume a DMA transfer. Even worse,
 * it seems that the DI or DO bit in ISR1 doesn't get raised after a DMA
 * transfer when the device wasn't re-addressed. Thus transfers that have
 * to be split in several chunks due to the transfer buffer not being
 * large enough can only be done using PIO.
 *-------------------------------------------------------------------------*/

int board_read( Board_t * board )
{
	if ( board->status & USE_DMA && board->transf.cnt > 1 )
		return board_dma_read( board );
	else
		return board_pio_read( board );
}


/*----------------------------------------------------------------*
 * The function reads in using DMA as many bytes as fit into the
 * transfer buffer in the board structure or until an END signal
 * was received. The size of the transfer buffer as well as the
 * information if this is the last chunk of data to be read in a
 * series of read opreations. The function expects the device to
 * addressed as talker and the board as listener and the ATN line
 * to be unasserted.
 *----------------------------------------------------------------*/

static int board_dma_read( Board_t * board )
{
	int ccount;


	PDEBUGin( );

	/* Clear ISR1 register and the FIFO */

	board_in( reg_addr.ISR0 );
	board_in( reg_addr.ISR1 );
	board_in( reg_addr.ISR2 );
	board_out( reg_addr.CMDR, RESET_FIFO );

	/* Setup board to read data from the bus in 16 bit transfer mode. */

	regs.cfg |= CFG_IN | CFG_TLCHLTE | CFG_16_8N;
	regs.cfg &= ~ CFG_CMD;
	board_out( reg_addr.CFG, regs.cfg );

	/* Write maximum number of bytes to be read, i.e. the length of the
	   data buffer into the count registers */

	ccount = - board->transf.cnt;
	board_out( reg_addr.CNT0,
		   ( unsigned char ) ( ccount & 0xFF ) );
	board_out( reg_addr.CNT1,
		   ( unsigned char ) ( ( ccount >>  8 ) & 0xFF ) );
	board_out( reg_addr.CNT2,
		   ( unsigned char ) ( ( ccount >> 16 ) & 0xFF ) );
	board_out( reg_addr.CNT3,
		   ( unsigned char ) ( ( ccount >> 24 ) & 0xFF ) );

	/* Enable board side DMA */

	if ( ! ( regs.accwr & ACCWR_DMAEN ) ) {
		regs.accwr |= ACCWR_DMAEN;
		board_out( reg_addr.ACCWR, regs.accwr );
	}

	/* Clear IMR0 and IMR2 */

	if ( regs.imr0 & 0x7F ) {
		regs.imr0 = 0x80;
		board_out( reg_addr.IMR0, regs.imr0 );
	}

	if ( regs.imr2 ) {
		regs.imr2 = 0;
		board_out( reg_addr.IMR2, regs.imr2 );
	}

	/* Read in all bytes via the FIFO using 16-bit DMA mode (take care:
	   the second argument is the maximum number of expected bytes and
	   must be an even number - so better tell the DMA chip that we're
	   going to send one more byte than we really will read instead of
	   sending it an odd value!) */

	gpib_isa_dma_enable( board->transf.buf,
			     ( board->transf.cnt + 1 ) & 0xfffe, READ );

	/* Enable interrupts and make board flag errors and DONE state.
	   Then set up the transfer semaphore and switch on interrupts */

	semaphore_init( &board->transf.sem, 0 );

	regs.imr1 |= IMR1_END_IE;
	board_out( reg_addr.IMR1, regs.imr1 );

	regs.imr3 |= IMR3_TLCINT_IE | IMR3_DONE_IE;
	board_out( reg_addr.IMR3, regs.imr3 );

	regs.intr |= INTR_INTEN;
	board_out( reg_addr.INTR, regs.intr );

	/* Send the GO command */

	board_out( reg_addr.CMDR, GO );

	/* Wait for transfer to finish (or interrupt or timeout to occur) */

	if ( ! board->timeout )
		down( &board->transf.sem );

	/* Disable host side DMA */

	gpib_isa_dma_disable( );

	/* Send STOP command and reset command bit in CFG register */

	board_out( reg_addr.CMDR, STOP );

	/* Disable DMA and interrupts */

	if ( regs.accwr & ACCWR_DMAEN ) {
		regs.accwr &= ~ ACCWR_DMAEN;
		board_out( reg_addr.ACCWR, regs.accwr );
	}

	regs.intr &= ~ INTR_INTEN;
	board_out( reg_addr.INTR, regs.intr );

	regs.imr3 &= ~ ( IMR3_TLCINT_IE | IMR3_DONE_IE );
	board_out( reg_addr.IMR3, regs.imr3 );

	/* Reset CFG and ISR1 register to normal state */

	regs.cfg &= ~ ( CFG_IN | CFG_TLCHLTE );
	board_out( reg_addr.CFG, regs.cfg );

	regs.imr1 &= ~ IMR1_END_IE;
	board_out( reg_addr.IMR1, regs.imr1 );

	/* Calculate how many bytes have been read */

	ccount = (   ( board_in( reg_addr.CNT0 ) & 0xFF )	     |
		   ( ( board_in( reg_addr.CNT1 ) & 0xFF ) <<  8 ) |
		   ( ( board_in( reg_addr.CNT2 ) & 0xFF ) << 16 ) |
		   ( ( board_in( reg_addr.CNT3 ) & 0xFF ) << 24 ) );
	gpib_count = board->transf.cnt + ccount;

	PDEBUG( "Read %ld bytes\n", gpib_count );

	/* Set the result flags */

	if ( board->timeout ) {
		gpib_status |= GPIB_ERR | GPIB_TIMO;
		gpib_error   = GPIB_EABO;
	} else if ( board_in( reg_addr.ISR1 ) & ISR1_END_RX )
		gpib_status |= GPIB_END;

	PDEBUGout( );
	return gpib_status;
}


/*---------------------------------------------------------------------*/
/* The function reads in as many bytes as fit into the transfer buffer */
/* in the board structure or until an END signal was received using    */
/* PIO. The size of the transfer buffer as well as the information if  */
/* this is the last chunk of data to be read in a series of read       */
/* operations. The function expects the device to be addressed as      */
/* talker and the board as listener and the ATN line to be unasserted. */
/*---------------------------------------------------------------------*/

static int board_pio_read( Board_t * board )
{
	unsigned int count = 0;
	int ccount;


	PDEBUGin( );

	/* Clear ISR1 register and the FIFO */

	board_in( reg_addr.ISR0 );
	board_in( reg_addr.ISR1 );
	board_in( reg_addr.ISR2 );
	board_out( reg_addr.CMDR, RESET_FIFO );

	/* Setup board to read data from the bus in 8 bit transfer mode.
	   Reads only happen through FIFO B. */

	regs.cfg |= CFG_IN | CFG_TLCHLTE;
	regs.cfg &= ~ ( CFG_CMD | CFG_16_8N | CFG_A_BN );
	board_out( reg_addr.CFG, regs.cfg );

	/* Write maximum number of bytes to be read, i.e. the length of the
	   data buffer into the count registers */

	ccount = - board->transf.cnt;
	board_out( reg_addr.CNT0,
		   ( unsigned char ) ( ccount & 0xFF ) );
	board_out( reg_addr.CNT1,
		   ( unsigned char ) ( ( ccount >>  8 ) & 0xFF ) );
	board_out( reg_addr.CNT2,
		   ( unsigned char ) ( ( ccount >> 16 ) & 0xFF ) );
	board_out( reg_addr.CNT3,
		   ( unsigned char ) ( ( ccount >> 24 ) & 0xFF ) );

	/* Make sure no DMA is used and clear IMR0 and IMR2 */

	if ( regs.accwr & ACCWR_DMAEN ) {
		regs.accwr &= ~ ACCWR_DMAEN;
		board_out( reg_addr.ACCWR, regs.accwr );
	}

	if ( regs.imr0 & 0x7F) {
		regs.imr0 = 0x80;
		board_out( reg_addr.IMR0, regs.imr0 );
	}

	if ( regs.imr2 ) {
		regs.imr2 = 0;
		board_out( reg_addr.IMR2, regs.imr2 );
	}

	/* Enable interrupts (if they're to be used) and make board flag
	   errors, FIFO not empty and DONE. Then set up the transfer
	   semaphore and switch on interrupts */

	if ( board->status & USE_IRQ ) {
		semaphore_init( &board->transf.sem, 0 );

		regs.imr1 |= IMR1_END_IE;
		board_out( reg_addr.IMR1, regs.imr1 );

		regs.imr3 |= IMR3_TLCINT_IE | IMR3_DONE_IE;
		board_out( reg_addr.IMR3, regs.imr3 );

		regs.intr |= INTR_INTEN;
		board_out( reg_addr.INTR, regs.intr );
	}

	/* Send the GO command */

	board_out( reg_addr.CMDR, GO );

	/* Read in all bytes via FIFO using 8-bit PIO mode */

	end = done = 0;

	PDEBUG( "Starting PIO read loop\n" );

	while ( count < board->transf.cnt ) {
		if ( ! ( board_in( reg_addr.ISR3 ) & ISR3_NEF ) ) {
			board_wait_in( board );
			if ( board->timeout )
				goto rd_done;
		}

		* ( board->transf.buf + count++ ) = board_in( reg_addr.FIFOB );

		if ( done || end || board->timeout )
			goto rd_done;
	}

	/* Polled wait for end of transfer when everything was read */

	while ( ! ( done || board->timeout ) ) {
		done = board_in( reg_addr.ISR3 ) & ISR3_DONE;
		if ( ! end )
			end = board_in( reg_addr.ISR1 ) & ISR1_END_RX;
	}

 rd_done:

	/* Send STOP command and reset command bit in CFG register */

	board_out( reg_addr.CMDR, STOP );

	/* Disable the interrupts (if used) */

	if ( board->status & USE_IRQ ) {
		regs.intr &= ~ INTR_INTEN;
		board_out( reg_addr.INTR, regs.intr );

		regs.imr3 &= ~ ( IMR3_TLCINT_IE | IMR3_DONE_IE );
		board_out( reg_addr.IMR3, regs.imr3 );
	}

	/* Reset CFG and ISR1 register to normal state */

	regs.cfg &= ~ ( CFG_IN | CFG_TLCHLTE );
	board_out( reg_addr.CFG, regs.cfg );

	regs.imr1 &= ~ IMR1_END_IE;
	board_out( reg_addr.IMR1, regs.imr1 );

	/* Calculate how many bytes have been read from the bus */

	ccount = (   ( board_in( reg_addr.CNT0 ) & 0xFF )	     |
		   ( ( board_in( reg_addr.CNT1 ) & 0xFF ) <<  8 ) |
		   ( ( board_in( reg_addr.CNT2 ) & 0xFF ) << 16 ) |
		   ( ( board_in( reg_addr.CNT3 ) & 0xFF ) << 24 ) );
	gpib_count = board->transf.cnt + ccount;

	PDEBUG( "Got %ld byte from device\n", gpib_count );

	/* Set the result flags */

	if ( board->timeout ) {
		gpib_status |= GPIB_ERR | GPIB_TIMO;
		gpib_error   = GPIB_EABO;
	} else if ( end || board_in( reg_addr.ISR1 ) & ISR1_END_RX )
		gpib_status |= GPIB_END;

	PDEBUGout( );
	return gpib_status;
}


/*
 * Local variables:
 * c-basic-offset: 8
 * c-indent-level: 8
 * c-brace-imaginary-offset: 0
 * c-brace-offset: 0
 * c-argdecl-indent: 4
 * c-label-offset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: t
 * tab-width: 8
 * End:
 */
