/*
  $Id: cmd.c 714 2006-03-07 10:21:18Z jens $

  Copyright (C) 2001-2006 Jens Thoms Toerring
 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
  To contact the author send email to: jt@toerring.de
*/


#include "gpib_module.h"
#include "sys.h"
#include "board.h"
#include "pcII.h"


static unsigned char board_wait_cmd( Board_t * board );


/*------------------------------------------------------------------*
 * Function writes a command to the bus. It must only be called if
 * the board is the ACIC, i.e. it's the controller and the ATN line
 * is raised!
 *------------------------------------------------------------------*/

int board_cmd( Board_t * board )
{
	unsigned char err = 0;
	bool stop_timer = 0;


	PDEBUGin( );

	/* Enable interrupt (if to be used) - if the CO bit in ISR2 is already
	   set we must initialize the transfer semaphore not to lock
	   immediately (otherwise the first call of board_wait_cmd() would
	   freeze everything even though we're allowed to write) */

	if ( board->status & USE_IRQ ) {
		semaphore_init( &board->transf.sem,
			       ( board_in( reg_addr.ISR2 ) & HR_CO ) ? 1 : 0 );
		if ( board->type == PCIIA )
			board_out( ( void * ) ( 0x2f0 | board->irq ), 0xff );
		board_out( reg_addr.IMR2, HR_COIE );
		board_out( reg_addr.IMR1, HR_ERR );
	}

	/* Start the timer (unless it's already running or there was a
	   timeout) */

	if ( ! ( board->status & TIMER_ON || board->timeout ) ) {
		gpib_start_timer( );
		stop_timer = SET;
	}

	/* Write out all bytes to the bus using PIO mode (DMA doesn't work
	   with commands because the DMA control lines only influence the
	   DI and DO bits in ISR1 but not the CO bit in ISR2) */

	while ( gpib_count < board->transf.cnt ) {
		if ( ( err = board_wait_cmd( board ) ) || board->timeout )
			break;
		board_out( reg_addr.CDOR,
			   * ( unsigned char * )
			   ( board->transf.buf + gpib_count++ ) );
	}

	if ( ! ( err || board->timeout ) )
		err = board_wait_cmd( board );

	/* Stop timer if necessary */

	if ( stop_timer )
		gpib_stop_timer( );

	/* Disable the interrupt (if it was used at all) */

	if ( board->status & USE_IRQ )
	{
		board_out( reg_addr.IMR2, 0 );
		board_out( reg_addr.IMR1, 0 );
	}

	/* Set the result flags */

	if ( err ) {
		gpib_status |= GPIB_ERR;
		if ( board->timeout )
			gpib_status |= GPIB_TIMO;
		gpib_error   = GPIB_ENOL;
		PDEBUG( "No listener in cmd\n" );
	} else if ( board->timeout ) {
		gpib_status |= GPIB_ERR | GPIB_TIMO;
		gpib_error   = GPIB_EABO;
		PDEBUG( "Timeout in cmd\n" );
	} else if ( board->transf.is_last && gpib_count == board->transf.cnt )
		gpib_status |= GPIB_CMPL;

	PDEBUGout( );
	return gpib_status;
}


/*-----------------------------------------------------------------*
 * When using the interrupt do a down on the transfer semaphore -
 * it will be up'ed either by the interrupt handler or on timeout.
 * Without interrupt just poll the ISR2 register until CDOR can
 * accept a new command byte (CO in ISR2 is set), or ERR in ISR1
 * is set (indicating an error) or a timeout occurs.
 *-----------------------------------------------------------------*/

static unsigned char board_wait_cmd( Board_t * board )
{
	unsigned char isr1 = 0;


	PDEBUGin( );

	if ( board->status & USE_IRQ )
	{
		/* Don't call down() if a timeout or error already occurred -
		   a timeout may already have happend before the semaphore
		   got initialized! */

		if ( ! ( ( isr1 = board_in( reg_addr.ISR1 ) ) & HR_ERR ||
			 board->timeout ) ) {
			down( &board->transf.sem );
			isr1 = board_in( reg_addr.ISR1 );
		}
		if ( board->type == PCIIA )
			board_out( ( void * ) ( 0x2f0 | board->irq ), 0xff );
	} else
		while ( ! ( isr1 = board_in( reg_addr.ISR1 ) & HR_ERR ) &&
			! ( board_in( reg_addr.ISR2 ) & HR_CO ) &&
			! board->timeout )
			/* empty */ ;

	PDEBUGout( );
	return ( isr1 & HR_ERR ) ? 1 : 0;
}


/*
 * Local variables:
 * c-basic-offset: 8
 * c-indent-level: 8
 * c-brace-imaginary-offset: 0
 * c-brace-offset: 0
 * c-argdecl-indent: 4
 * c-label-ofset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: t
 * tab-width: 8
 * End:
 */
