/*
  $Id: init.c 732 2006-05-30 20:13:39Z jens $

  Copyright (C) 2001-2006 Jens Thoms Toerring
 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
  To contact the author send email to: jt@toerring.de
*/


#include "gpib_module.h"
#include "autoconf.h"
#include "sys.h"
#include "board.h"
#include "pcII.h"


unsigned char err;
unsigned char done;
PCII_Register_Adresses_t reg_addr;


static bool board_pcII_request_region( Board_t * board );
static bool board_pcIIa_request_region( Board_t * board );
static void board_pcII_release_region( Board_t * board );
static void board_pcIIa_release_region( Board_t * board );
static bool board_init_chip( Board_t * board );


/*-----------------------------------------------------------------*
 * Does everything to initialize the board: The I/O region for the
 * board is locked and the chip is initialized.
 *-----------------------------------------------------------------*/

bool board_init( Board_t * board )
{
	PDEBUGin( );

	/* Store the board type */

#if defined GPIB_PCII
	board->type = PCII;
#else
	board->type = PCIIA;
#endif

	/* Get a lock on the memory locations used by the board and set up the
	   register pointers to these locations */ 

	if ( ( board->type == PCII && ! board_pcII_request_region( board ) ) ||
	     ( board->type == PCIIA &&
	       ! board_pcIIa_request_region( board ) ) ) {
		PDEBUGout( );
		return FAIL;
	}

	/* Initialize the chip */

	if ( ! board_init_chip( board ) ) {
		if ( board->type == PCII )
			board_pcII_release_region( board );
		if ( board->type == PCIIA )
			board_pcIIa_release_region( board );
		PDEBUGout( );
		return FAIL;
	}

	/* Set EOS character and eosmode (the values to be set have both
	   already been set up) */

	board_set_eos( board );
	board_set_eosmode( board, HR_HLDA );
	board->status |= HOLDOFF;

	PDEBUGout( );
	return OK;
}


/*-----------------------------------------------------------------------*
 * Called when the module is becoming unloaded: the board goes to Idle
 * Controller State and the I/O region locked for the board is released.
 *-----------------------------------------------------------------------*/

void board_cleanup( Board_t * board )
{
	PDEBUGin( );

	/* Make board to Idle Controller state */

	board_release_control( board );

	/* Release I/O region */

	if ( board->type == PCII )
		board_pcII_release_region( board );

	if ( board->type == PCIIA )
		board_pcIIa_release_region( board );

	PDEBUGout( );
}


/*--------------------------------------------------*
 * Requests a fitting I/O region for the pcII board
 *--------------------------------------------------*/

static bool board_pcII_request_region( Board_t * board )
{
	PORT base;


	PDEBUGin( );

	/* The board has an I/O address range of 8 bytes */

	board->io_size = 8;

	/* Check that the boards base address is valid */

	for ( base = 0x0100; base <= 0x03f8; base += board->io_size )
		if ( base == ( PORT ) board->base )
			break;

	if ( base > 0x0400 )
	{
		printk( KERN_ERR "%s: Invalid base address %p for pcII "
			"board.\n", board->name, board->base );
		return FAIL;
	}

	/* Check that the 8 byte at the boards address are not already
	   in use */

#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 6, 0 )
	if ( request_region( base, board->io_size, board->name ) == NULL ) {
		printk( KERN_ERR "%s: Can't get region at boards base "
			"address %p\n", board->name, board->base );
		return FAIL;
	}
#else
	if ( check_region( base, board->io_size ) < 0 ) {
		printk( KERN_ERR "%s: Can't get region at boards base "
			"address %p\n", board->name, board->base );
		PDEBUGout( );
		return FAIL;
	}

	/* Now lock this region - this always succeeds when the check
           succeeded */

	request_region( base, board->io_size, board->name );
#endif

	/* Set up the pointers to the memory locations associated with the
	   chips registers */

	reg_addr.DIR  = ( unsigned char * ) base++;
	reg_addr.ISR1 = ( unsigned char * ) base++;
	reg_addr.ISR2 = ( unsigned char * ) base++;
	reg_addr.SPSR = ( unsigned char * ) base++;
	reg_addr.ADSR = ( unsigned char * ) base++;
	reg_addr.CPTR = ( unsigned char * ) base++;
	reg_addr.ADR0 = ( unsigned char * ) base++;
	reg_addr.ADR1 = ( unsigned char * ) base;

	PDEBUGout( );
	return OK;
}


/*---------------------------------------------------*
 * Requests a fitting I/O region for the pcIIa board
 *---------------------------------------------------*/

static bool board_pcIIa_request_region( Board_t * board )
{
	PORT base;
	int i;


	PDEBUGin( );

	/* Check that the boards base address is valid */

	for ( base = 0x02e1; base <= 0x62e1; base += 0x2000 )
		if ( base == ( PORT ) board->base )
			break;

	if ( base > 0x62e1 ) {
		printk( KERN_ERR "%s: Invalid base address %p for pcIIa "
			"board.\n", board->name, board->base );
		PDEBUGout( );
		return FAIL;
	}

#if LINUX_VERSION_CODE < KERNEL_VERSION( 2, 6, 0 )

	/* Check that the memory addresses used by the board are not already
	   in use (eventhough the upper 8 locations are not going to be used
	   they have to be checked and locked to avoid other devices getting
	   them) */

	for ( i = 0; i < 8; i++ )
		if ( check_region( base + i * 0x400, 1 ) < 0 ||
		     check_region( base + 0x8000 + i * 0x400, 1 ) < 0 ) {
			printk( KERN_ERR "%s: Can't lock region at boards "
				"base address %p\n",
				board->name, board->base );
			PDEBUGout( );
			return FAIL;
		}

	/* At least the NI board with a NAT4882 in 7210 emulation mode seems
	   to need the port at ( 0x2f0 | board->irq ) when using interrupts. I
	   didn't figure out yet what this is needed for but it won't work
	   without... */

	if ( board->use_irq && ( board->irq >= 2 || board->irq <= 7 ) &&
	     check_region( 0x2f0 | board->irq, 1 ) ) {
		printk( KERN_ERR "%s: Can't lock region for interrupt "
			"handling at %p\n",
			board->name, ( void * ) ( 0x2f0 | board->irq ) );
		PDEBUGout( );
		return FAIL;
	}
#endif

	/* Now lock these memory locations */

#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 6, 0 )
	for ( i = 0; i < 8; i++ )
		if ( request_region( base + i * 0x400, 1, board->name ) )
			break;

	if ( i < 8 ) {
		for ( ; i >= 0; i-- )
			release_region( base + i * 0x400, 1 );

		printk( KERN_ERR "%s: Can't lock region at boards base "
			"address %p\n", board->name, board->base );
		PDEBUGout( );
		return FAIL;
	}

	for ( i = 0; i < 8; i++ )
		if ( request_region( base + 0x8000 + i * 0x400, 1,
				     board->name ) )
			break;

	if ( i < 8 ) {
		for ( ; i >= 0; i-- )
			release_region( base + 0x8000 + i * 0x400, 1 );

		for ( i = 0; i < 8; i++ ) 
			release_region( base + 0x400 * i, 1 );

		printk( KERN_ERR "%s: Can't lock region at boards base "
			"address %p\n", board->name, board->base );
		PDEBUGout( );
		return FAIL;
	}

	if ( board->use_irq && ( board->irq >= 2 || board->irq <= 7 ) &&
	     request_region( 0x2f0 | board->irq, 1, board->name ) ) {
		for ( i = 0; i < 8; i++ ) {
			release_region( base + 0x400 * i, 1 );
			release_region( base + 0x8000 + i * 0x400, 1 );
		}

		printk( KERN_ERR "%s: Can't lock region for interrupt "
			"handling at %p\n",
			board->name, ( void * ) ( 0x2f0 | board->irq ) );
		PDEBUGout( );
		return FAIL;
	}
#else
	for ( i = 0; i < 8; i++ ) {
		request_region( base + i * 0x400, 1, board->name );
		request_region( base + 0x8000 + i * 0x400, 1, board->name );
	}

	if ( board->use_irq && ( board->irq >= 2 || board->irq <= 7 ) )
		request_region( 0x2f0 | board->irq, 1, board->name );
#endif

	/* Set up the pointers to the memory locations associated with the
	   chips registers */

	reg_addr.DIR  = ( unsigned char * ) base;
	base += 0x400;
	reg_addr.ISR1 = ( unsigned char * ) base;
	base += 0x400;
	reg_addr.ISR2 = ( unsigned char * ) base;
	base += 0x400;
	reg_addr.SPSR = ( unsigned char * ) base;
	base += 0x400;
	reg_addr.ADSR = ( unsigned char * ) base;
	base += 0x400;
	reg_addr.CPTR = ( unsigned char * ) base;
	base += 0x400;
	reg_addr.ADR0 = ( unsigned char * ) base;
	base += 0x400;
	reg_addr.ADR1 = ( unsigned char * ) base;

	PDEBUGout( );
	return OK;
}


/*-------------------------------------------------*
 * Releases the I/O region locked for a pcII board
 *-------------------------------------------------*/

static void board_pcII_release_region( Board_t * board )
{
	PDEBUGin( );
	release_region( ( PORT ) board->base, board->io_size );
	PDEBUGout( );
}


/*--------------------------------------------------*
 * Releases the I/O region locked for a pcIIa board
 *--------------------------------------------------*/

static void board_pcIIa_release_region( Board_t * board )
{
	int i;


	PDEBUGin( );

	for ( i = 0; i < 8; i++ ) {
		release_region( ( PORT ) ( board->base + 0x400 * i ), 1 );
		release_region( ( PORT ) ( board->base + 0x8000 + 0x400 * i ),
				1 );
	}

	if ( board->use_irq && ( board->irq >= 2 && board->irq <= 7 ) )
		release_region( ( PORT ) (0x2f0 | board->irq ), 1 );

	PDEBUGout( );
}


/*-----------------------------------------------------------------*
 *-----------------------------------------------------------------*/

static bool board_init_chip( Board_t * board )
{
	PDEBUGin( );

	/* Place the chip in a known quiescent state by writing the chip reset
	   auxiliary command to the AUXMR */

	board_out( reg_addr.AUXMR, AUX_cr );

	/* For a start clear both interrupt mask registers */

	board_out( reg_addr.IMR1, 0 );
	board_out( reg_addr.IMR2, 0 );

	/* Before setting the final address of the board we use the address
	   register 0 to figure out if there's really a chip at all - if
	   setting an address and then reading it fails we take this as an
	   indication that there is no chip at all. */

	board_out( reg_addr.ADR, HR_DL | HR_DT | 0x15 );
	if ( ( board_in( reg_addr.ADR0 ) & 0x1F ) != 0x15 )
	{
		printk( KERN_ERR "%s: Failed to initialize chip, not a "
			"uPD7210.\n", board->name );
		PDEBUGout( );
		return FAIL;
	}

	board_out( reg_addr.ADR, HR_DL | HR_DT | 0x0A );
	if ( ( board_in( reg_addr.ADR0 ) & 0x1F ) != 0x0A )
	{
		printk( KERN_ERR "%s: Failed to initialize chip, not a "
			"uPD7210.\n", board->name );
		PDEBUGout( );
		return FAIL;
	}

	board_out( reg_addr.ADR, HR_ARS | HR_DL | HR_DT | 0x15 );
	if ( ( board_in( reg_addr.ADR1 ) & 0x1F ) != 0x15 )
	{
		printk( KERN_ERR "%s: Failed to initialize chip, not a "
			"uPD7210.\n", board->name );
		PDEBUGout( );
		return FAIL;
	}

	board_out( reg_addr.ADR, HR_ARS | HR_DL | HR_DT | 0x0A );
	if ( ( board_in( reg_addr.ADR1 ) & 0x1F ) != 0x0A )
	{
		printk( KERN_ERR "%s: Failed to initialize chip, not a "
			"uPD7210.\n", board->name );
		PDEBUGout( );
		return FAIL;
	}

	/* Set the primary (and secondary if given) address of the board. If a
	   secondary address is specified use address mode 2, otherwise use
	   address mode 1 and completely disable the secondary address */

	board_out( reg_addr.ADR, board->pad & 0x1F );

	if ( board->sad >= 0x60 && board->sad <= 0x7E ) {
		board_out( reg_addr.ADMR, HR_TRM1 | HR_TRM0 | HR_ADM1 );
		board_out( reg_addr.ADR, HR_ARS | ( board->sad & 0x1F ) );
	} else {
		board_out( reg_addr.ADMR, HR_TRM1 | HR_TRM0 | HR_ADM0 );
		board_out( reg_addr.ADR, HR_ARS | HR_DT | HR_DL );
	}

	/* Clear the serial poll mask register */

	board_out( reg_addr.SPMR, 0 );

	/* Clear the the parallel poll register */

	board_out( reg_addr.AUXMR, PPR | 0 );

	/* Release pon state that was entered due to the chip reset */

	board_out( reg_addr.AUXMR, AUX_pon );

	PDEBUGout( );
	return OK;
}


/*-------------------------------------------------------------------*
 * Makes the board release its control over the GPIB bus by going to
 * the Idle Controller state. If it's alredy in this state calling
 * this function does nothing.
 *-------------------------------------------------------------------*/

int board_release_control( Board_t * board )
{
	PDEBUGin( );

	/* If the board is still CIC send the local chip reset message, that
	   should do the job in any case - but afterwards also we seem to have
	   to send te pon message, otherwise the board doesn't like to get
	   back online later on... */

	if ( ( board_in( reg_addr.ADSR ) & HR_CIC ) ) {
		board_out( reg_addr.AUXMR, AUX_cr );
		board_out( reg_addr.AUXMR, AUX_pon );
	}

	gpib_status = board_get_state( board );

	PDEBUGout( );
	return gpib_status;
}


/*
 * Local variables:
 * c-basic-offset: 8
 * c-indent-level: 8
 * c-brace-imaginary-offset: 0
 * c-brace-offset: 0
 * c-argdecl-indent: 4
 * c-label-ofset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: t
 * tab-width: 8
 * End:
 */
