/*
  $Id: util.c 717 2006-03-09 13:25:06Z jens $

  Copyright (C) 2001-2006 Jens Thoms Toerring
 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
  To contact the author send email to: jt@toerring.de
*/


#include "gpib_module.h"
#include "sys.h"
#include "board.h"
#include "pcII.h"


static int board_wait_polled( Board_t * board,
			      int       flags );
static int board_wait_irq( Board_t * board,
			   int       flags );


/*------------------------------------------------------------------*
 * Outputs one byte to the register addressed by the first argument
 *------------------------------------------------------------------*/

void board_out( unsigned char * addr,
		unsigned char   byte )
{
#ifdef GPIB_PCIIA_SLOWDOWN
	udelay(100);            /* jg - needed for rev A PCIIa card */
#endif
	outb_p( byte, ( PORT ) addr );
}


/*-------------------------------------------------------------*
 * Inputs one byte from the register addressed by the argument
 *-------------------------------------------------------------*/

unsigned char board_in( unsigned char * addr )
{
#ifdef GPIB_PCIIA_SLOWDOWN
	udelay(100);            /* jg - needed for rev A PCIIa card */
#endif
	return inb_p( ( PORT ) addr );
}


/*------------------------------------------------------------*
 * Function reads in the address status register and sets the
 * CIC, ATN, LACS and TACS bits in gpib_status accordingly.
 *------------------------------------------------------------*/

int board_get_state( Board_t * board )
{
	unsigned char state;

	PDEBUGin( );

	state = board_in( reg_addr.ADSR );

	if ( state & HR_CIC )
		gpib_status |= GPIB_CIC;
	else
		gpib_status &= ~ GPIB_CIC;

	if ( state & HR_NATN )
		gpib_status &= ~ GPIB_ATN;
	else
		gpib_status |= GPIB_ATN;

	if ( state & HR_LA )
		gpib_status |= GPIB_LACS;
	else
		gpib_status &= ~ GPIB_LACS;

	if ( state & HR_TA )
		gpib_status |= GPIB_TACS;
	else
		gpib_status &= ~ GPIB_TACS;

	PDEBUG( "gpib_status = 0x%x\n", gpib_status );
	PDEBUGout( );
	return gpib_status;
}


/*-----------------------------------------------*
 * Function waits for either a SRQ from a device
 * and/or for an END condition.
 *-----------------------------------------------*/

int board_wait( Board_t * board,
		int       flags )
{
	int state;


	PDEBUGin( );

	if ( flags & GPIB_END )
		gpib_status &= ~GPIB_END;
	if ( flags & GPIB_RQS )
		gpib_status &= ~GPIB_RQS;

	if ( board->status & USE_IRQ )
		state = board_wait_irq( board, flags );
	else
		state = board_wait_polled( board, flags );

	PDEBUGout( );
	return gpib_status |= state;
}


/*------------------------------------------------------*
 * Waits for SRQ or END event by constantly polling the
 * necessary registers.
 *------------------------------------------------------*/

static int board_wait_polled( Board_t * board,
			      int       flags )
{
	int state = 0;


	while ( ! board->timeout && ! state )
	{
		if ( flags & GPIB_END &&
		     board_in( reg_addr.ISR1 ) & HR_END )
			state |= GPIB_END;
		if ( flags & GPIB_END &&
		     board_in( reg_addr.ISR2 ) & HR_SRQI )
			state |= GPIB_RQS;
	}

	return state;
}


/*--------------------------------------------------------*
 * Waits for SRQ or END event by waiting for an interrupt
 * generated by these events.
 *--------------------------------------------------------*/

static int board_wait_irq( Board_t * board,
			   int       flags )
{
	int state = 0;


	/* First check if one of the events we're waiting for hasn't
	   already happend */

	if ( flags & GPIB_END &&
	     board_in( reg_addr.ISR1 ) & HR_END )
		state |= GPIB_END;
	if ( flags & GPIB_END &&
	     board_in( reg_addr.ISR2 ) & HR_SRQI )
		state |= GPIB_RQS;

	if ( state )
		return state;

	semaphore_init( &board->transf.sem, 0 );

	if ( board->type == PCIIA )
		board_out( ( void * ) ( 0x2f0 | board->irq ), 0xff );
	if ( flags & GPIB_END )
		board_out( reg_addr.IMR1, HR_ENDIE );
	if ( flags & GPIB_RQS )
		board_out( reg_addr.IMR2, HR_SRQIE );

	/* Wait for interrupt to happen (or a timeout) */

	if ( ! board->timeout )
		down( &board->transf.sem );

	/* Disable further interrupts */

	board_out( reg_addr.IMR1, 0 );
	board_out( reg_addr.IMR2, 0 );
	if ( board->type == PCIIA )
		board_out( ( void * ) ( 0x2f0 | board->irq ), 0xff );

	if ( flags & GPIB_END &&
	     board_in( reg_addr.ISR1 ) & HR_END )
		state |= GPIB_END;
	if ( flags & GPIB_END &&
	     board_in( reg_addr.ISR2 ) & HR_SRQI )
		state |= GPIB_RQS;

	return state;
}


/*--------------------------------------------------------*
 *--------------------------------------------------------*/

int board_rpp( Board_t * board,
	       char *    c )
{
	PDEBUGin( );

	/* This function requires the board to be CIC */

	if ( ! ( gpib_status & GPIB_CIC ) ) {
		gpib_status |= GPIB_ERR;
		gpib_error   = GPIB_ECIC;
		PDEBUG( "Board not CIC\n" );
		PDEBUGout( );
		return gpib_status;
	}

	/* The board must also be active controller */

	if ( ! ( gpib_status & GPIB_ATN ) && board_tca( board ) & GPIB_ERR ) {
		gpib_status |= GPIB_ERR;
		gpib_error   = GPIB_ECIC;
		PDEBUG( "Can't make board ACIC\n" );
		PDEBUGout( );
		return gpib_status;
	}

	/* Do the parallel poll and read the reply from the devices */

	board_out( reg_addr.AUXMR, AUX_epp );
	udelay( 2 );
	*c = board_in( reg_addr.DIR );

	PDEBUGout( );

	return gpib_status;
}


/*--------------------------------------------------------------*
 * Raises the IFC line for (at least) 100 us and thus makes the
 * board CIC and also brings it into the active state.
 *--------------------------------------------------------------*/

int board_sic( Board_t * board )
{
	PDEBUGin( );

	/* Raise the IFC line for (at least) 100 us */

	board_out( reg_addr.AUXMR, AUX_sifc );
	udelay( GPIB_MIN_TIME );
	board_out( reg_addr.AUXMR, AUX_cifc );

	/* Wait for the board to become CIC and the ATN line to become set -
	   eventhough this should never fail it is probably better done
	   with the timer running to get us out of an infinite loop... */

	gpib_start_timer( );

	while ( ( board_get_state( board ) & GPIB_ACIC ) != GPIB_ACIC &&
		! board->timeout)
		/* empty */ ;

	if ( ! board->timeout )
		gpib_stop_timer( );
	else {
		gpib_status |= GPIB_ERR | GPIB_TIMO;
		gpib_error = GPIB_EABO;
		PDEBUG( "SIC failed, board not ACIC\n" );
	}

	PDEBUGout( );
	return gpib_status;
}


/*--------------------------------------------------------------*
 * Makes the board go from active to standby controller state.
 * This requires that the board is CIC and the ATN line is set.
 *--------------------------------------------------------------*/

int board_gts( Board_t * board )
{
	PDEBUGin( );

	/* Get the current board state */

	board_get_state( board );

	/* Check that board is CIC */

	if ( ! ( gpib_status & GPIB_CIC ) ) {
		gpib_status |= GPIB_ERR;
		gpib_error   = GPIB_ECIC;
		PDEBUG( "Board not CIC\n" );
		PDEBUGout( );
		return gpib_status;
	}

	/* Don't do anything if the ATN line isn't raised */

	if ( ! ( gpib_status & GPIB_ATN ) ) {
		PDEBUGout( );
		return gpib_status;
	}

	/* Send GTS auxiliary command and wait for ATN to become lowered -
	   eventhough this should never fail it is probably better done
	   with the timer running to get us out of an infinite loop... */

	gpib_start_timer( );

	board_out( reg_addr.AUXMR, AUX_gts );
	while ( board_get_state( board ) & GPIB_ATN && ! board->timeout )
		/* empty */ ;

	if ( ! board->timeout )
		gpib_stop_timer( );
	else {
		gpib_status |= GPIB_ERR | GPIB_TIMO;
		gpib_error = GPIB_EABO;
		PDEBUG( "GTS failed, board still ACIC\n" );
	}

	PDEBUGout( );
	return gpib_status;
}


/*--------------------------------------------*
 * Asserts or releases the REN line depending
 * on the value of the second argument.
 *--------------------------------------------*/

int board_ren( Board_t * board,
	       bool      remote )
{
	PDEBUGin( );

	/* Get the current board state */

	board_get_state( board );

	/* This function requires the board to be CIC */

	if ( ! ( gpib_status & GPIB_CIC ) ) {
		gpib_status |= GPIB_ERR;
		gpib_error   = GPIB_ECIC;
		PDEBUG( "Board not CIC\n" );
		PDEBUGout( );
		return gpib_status;
	}
		
	/* The board must even be active controller */

	if ( ! ( gpib_status & GPIB_ATN ) && board_tca( board ) & GPIB_ERR ) {
		gpib_status |= GPIB_ERR;
		gpib_error   = GPIB_ECIC;
		PDEBUG( "Can't make board ACIC\n" );
		PDEBUGout( );
		return gpib_status;
	}

	/* Now set or reset the REN line by the auxiliary command and make sure
	   it stays in this state for at least 100 us */

	if ( remote ) {
		board_out( reg_addr.AUXMR, AUX_sren );
		board->status |= REN;
	} else {
		board_out( reg_addr.AUXMR, AUX_cren );
		board->status &= ~REN;
	}

	udelay( GPIB_MIN_TIME );

	PDEBUGout( );
	return gpib_status;
}


/*------------------------------------------------*
 * Makes the board raise the ATN line immediately
 * and thereby become active controller.
 *------------------------------------------------*/

int board_tca( Board_t * board )
{
	PDEBUGin( );

	/* Get the current board state */

	board_get_state( board );

	/* Check that the board is CIC */

	if ( ! ( gpib_status & GPIB_CIC ) ) {
		gpib_status |= GPIB_ERR;
		gpib_error   = GPIB_ECIC;
		PDEBUG( "Board not CIC\n" );
		PDEBUGout( );
		return gpib_status;
	}

	/* Return if ATN line is already raised */

	if ( gpib_status & GPIB_ATN ) {
		PDEBUGout( );
		return gpib_status;
	}

	/* Take control asynchronously and wait for the ATN line to become set
	   - eventhough this should never fail it is probably better done with
	   the timer running to get us out of an infinite loop... */

	gpib_start_timer( );

	board_out( reg_addr.AUXMR, AUX_tca );
	while ( ! ( board_get_state( board ) & GPIB_ATN ) && ! board->timeout )
		/* empty */ ;

	if ( ! board->timeout )
		gpib_stop_timer( );
	else {
		gpib_status |= GPIB_ERR | GPIB_TIMO;
		gpib_error = GPIB_EABO;
		PDEBUG( "TCA failed, ATN not raised\n" );
	}

	PDEBUGout( );
	return gpib_status;
}


/*--------------------------------------------------------*
 * Makes the board raise the ATN line without corrupting
 * transferred data and thereby become active controller.
 *--------------------------------------------------------*/

int board_tcs( Board_t * board )
{
	PDEBUGin( );

	/* Get the current board state */

	board_get_state( board );

	/* Check that the board is CIC */

	if ( ! ( gpib_status & GPIB_CIC ) ) {
		gpib_status |= GPIB_ERR;
		gpib_error   = GPIB_ECIC;
		PDEBUG( "Board not CIC\n" );
		PDEBUGout( );
		return gpib_status;
	}

	/* Return if ATN line is already raised */

	if ( gpib_status & GPIB_ATN ) {
		PDEBUGout( );
		return gpib_status;
	}

	/* Take control asynchronously and wait for the ATN line to become set
	   - eventhough this should never fail it is probably better done with
	   the timer running to get us out of an infinite loop... */

	gpib_start_timer( );

	board_out( reg_addr.AUXMR, AUX_tcs );
	while ( ! ( board_get_state( board ) & GPIB_ATN ) && ! board->timeout )
		/* empty */ ;

	if ( ! board->timeout )
		gpib_stop_timer( );
	else {
		gpib_status |= GPIB_ERR | GPIB_TIMO;
		gpib_error = GPIB_EABO;
		PDEBUG( "TCA failed, ATN not raised\n" );
	}

	PDEBUGout( );
	return gpib_status;
}


/*------------------------------------------------------------------*
 * Function calculates the byte that has to be written to the board
 * to set the eos mode according to the eosmode entry of the board
 * structure.
 *------------------------------------------------------------------*/

void board_setup_eos_byte( Board_t * board )
{
	PDEBUGin( );
	board->eos_byte = AUXRA |
	                 ( ( board->eosmode & GPIB_BIN )  ? HR_BIN  : 0 ) |
		         ( ( board->eosmode & GPIB_REOS ) ? HR_REOS : 0 ) |
		         ( ( board->eosmode & GPIB_XEOS ) ? HR_XEOS : 0 );
	PDEBUGout( );
}


/*------------------------------------------------------*
 * Function sets the EOS character to be used according
 * to the eos entry in the board structure.
 *------------------------------------------------------*/

void board_set_eos( Board_t * board )
{
	PDEBUGin( );
	board_out( reg_addr.EOSR, board->eos );
	PDEBUGout( );
}


/*--------------------------------------------------------------*
 * Function sets the boards register for the eos mode according
 * to the eos_byte entry of the board structure, logically ored
 * with data passed to the function (this data should be either
 * HR_HLDE, HR_HLDA or 0 and nothing else, to make sure it gets
 * truncated to this values).
 *--------------------------------------------------------------*/

void board_set_eosmode( Board_t *     board,
			unsigned char v )
{
	PDEBUGin( );
	board_out( reg_addr.AUXMR,
		   board->eos_byte | ( v & ( HR_HLDE | HR_HLDA ) ) );
	PDEBUGout( );
}


/*-------------------------------------------------------------------*
 * Function sets a new primary address of the board. If the current
 * secondary address is invalid secondary addressing is switched off
 * and address mode 1 is used, otherwise it is switched on and
 * address mode 2 is used.
 *-------------------------------------------------------------------*/

void board_pad( Board_t * board )
{
	PDEBUGin( );

	board_out( reg_addr.ADR, board->pad & 0x1F );

	if ( board->sad >= 0x60 && board->sad <= 0x7e )
		board_out( reg_addr.ADMR, HR_TRM1 | HR_TRM0 | HR_ADM1 );
	else {
		board_out( reg_addr.ADMR, HR_TRM1 | HR_TRM0 | HR_ADM0 );
		board_out( reg_addr.ADR, HR_ARS | HR_DT | HR_DL );
	}

	PDEBUGout( );
}


/*-------------------------------------------------------------*
 * Function sets a new secondary address of the board. If the
 * address is invalid secondary addressing is switched off and
 * address mode 1 is used, otherwise it is switched on and
 * address mode 2 is used.
 *-------------------------------------------------------------*/

void board_sad( Board_t * board )
{
	PDEBUGin( );

	if ( board->sad >= 0x60 && board->sad <= 0x7e ) {
		board_out( reg_addr.ADMR, HR_TRM1 | HR_TRM0 | HR_ADM1 );
		board_out( reg_addr.ADR, HR_ARS | ( board->sad & 0x1F ) );
	} else {
		board_out( reg_addr.ADMR, HR_TRM1 | HR_TRM0 | HR_ADM0 );
		board_out( reg_addr.ADR, HR_ARS | HR_DT | HR_DL );
	}

	PDEBUGout( );
}


/*----------------------------------------------------------------------*
 * Because the board is not able to test the state of the control lines
 * directly the function returns only the states of the lines that can
 * be infered from other information...
 *----------------------------------------------------------------------*/

unsigned int board_lines( Board_t * board )
{
	unsigned int lines = 0;


	PDEBUGin( );
	
	lines |= ValidATN | ValidREN | ValidIFC;
	if ( gpib_status & GPIB_ATN )
		lines |= BusATN << 8;
	if ( board->status & REN )
		lines |= BusREN << 8;

	PDEBUGout( );
	return lines;
}


/*
 * Local variables:
 * c-basic-offset: 8
 * c-indent-level: 8
 * c-brace-imaginary-offset: 0
 * c-brace-offset: 0
 * c-argdecl-indent: 4
 * c-label-ofset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: t
 * tab-width: 8
 * End:
 */
