/*
  $Id: write.c 717 2006-03-09 13:25:06Z jens $

  Copyright (C) 2001-2006 Jens Thoms Toerring
 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
  To contact the author send email to: jt@toerring.de
*/

#include "gpib_module.h"
#include "sys.h"
#include "board.h"
#include "pcII.h"


static int board_dma_write( Board_t * board );
static int board_pio_write( Board_t * board );
static unsigned char board_wait_out( Board_t * board );


/*-------------------------------------------------------------------------*
 * Function decides if the transfer is to be done using DMA or PIO.
 * Unfortunately, it is not possible to resume a DMA transfer. Even worse,
 * it seems that the DI or DO bit in ISR1 doesn't get raised after a DMA
 * transfer when the device wasn't re-addressed. Thus transfers that have
 * to be split in several chunks due to the transfer buffer not being
 * large enough can only be done using PIO.
 *-------------------------------------------------------------------------*/

int board_write( Board_t * board )
{
	int retval;
	static bool cont = UNSET;             /* flag, set on split transfer */


	if ( ! board->transf.is_last )        /* is this a split transfer ? */
		cont = SET;

	/* DMA can only be used when it's allowed and for non-split
	   transfers */

	if ( board->status & USE_DMA && ! cont )
		retval = board_dma_write( board );
	else
		retval = board_pio_write( board );

	/* If this is the very end of a transfer reset 'cont' */

	if ( board->transf.is_last || gpib_status & GPIB_ERR )
		cont = UNSET;

	return retval;
}


/*-------------------------------------------------------------------*
 * Function writes out the bytes stored in the transfer buffer in
 * the board structure using DMA. Also the number of bytes is stored
 * in the board structure as well as the information if this is the
 * last of a series of write operations to the same device. The
 * function expects the device to be addressed as listener and the
 * board as talker and the ATN line to be unasserted.
 *-------------------------------------------------------------------*/

static int board_dma_write( Board_t * board )
{
	unsigned int send_count;
	int residue;
	unsigned char isr1 = 0;


	/* If there's only one byte PIO mode for output will do */

	if ( board->transf.cnt == 1 )
		return board_pio_write( board );

	PDEBUGin( );

	send_count = board->transf.cnt;

	/* If this is the last chunk to be written and we're supposed to raise
	   the EOI line with the last byte we need to keep one byte from being
	   written */

	if ( board->transf.is_last && board->eosmode & GPIB_EOT )
		send_count--;

	/* Enable DMA transfer */

	gpib_isa_dma_enable( board->transf.buf, send_count, WRITE );

	/* 'Arm' the transfer semaphore and enable interrupts and board side
	   DMA transfer. Asking for an interrupt on ENDIE is essential for
	   getting an interrupt when the DMA transfer is finished! */

	board_in( reg_addr.ISR1 );
	semaphore_init( &board->transf.sem, 0 );
	if ( board->type == PCIIA )
		board_out( ( void * ) ( 0x2f0 | board->irq ), 0xff );
	board_out( reg_addr.IMR1, HR_ENDIE | HR_ERRIE );
	board_out( reg_addr.IMR2, HR_DMAO );

	/* Wait for transfer to finish (or timeout to occur) */

	if ( ! board->timeout )
		down( &board->transf.sem );

	/* Disable interrupts and DMA transfer */

	board_out( reg_addr.IMR2, 0 );
	board_out( reg_addr.IMR1, 0 );

	/* Disable DMA transfer and check on success (in which case the
	   return value is zero) */

	residue = gpib_isa_dma_disable( );

	/* For the final byte of a transfer where the EOI line is required
	   to be set send this last byte now with PIO after raising the EOI
	   line. A bit hidden in here we also check in all cases (except
	   if a DMA error happened) that the last byte successfully arrived
	   at the device by testing the DO bit of ISR1. */

	board->status &= ~ USE_IRQ;

	if ( residue == 0 &&
	     ! ( isr1 = board_wait_out( board ) ) & HR_ERR &&
	     ! board->timeout &&
	     board->transf.is_last &&
	     board->eosmode & GPIB_EOT ) {
		board_out( reg_addr.AUXMR, AUX_seoi );
		board_out( reg_addr.CDOR, * ( unsigned char * )
					( board->transf.buf + send_count++ ) );
		if ( ! board->timeout )
			isr1 = board_wait_out( board );
		send_count++;
	}

	board->status |= USE_IRQ;

	/* Finally set the result flags etc. */

	gpib_count = send_count;

	if ( isr1 & HR_ERR ) {
		gpib_status |= GPIB_ERR;
		if ( board->timeout )
			gpib_status |= GPIB_TIMO;
		gpib_error   = GPIB_ENOL;
		PDEBUG( "No listener on DMA write\n" );
	} else if ( board->timeout ) {
		gpib_status |= GPIB_ERR | GPIB_TIMO;
		gpib_error   = GPIB_EABO;
		PDEBUG( "Timeout on DMA write\n" );
	} else if ( residue != 0 ) {           /* not everything got written */
		gpib_status |= GPIB_ERR;
		gpib_error   = GPIB_EDMA;
		PDEBUG( "DMA failure on write\n" );
	} else if ( board->transf.is_last )
		gpib_status |= GPIB_CMPL;

	PDEBUGout( );
	return gpib_status;
}


/*-------------------------------------------------------------------*
 * Function writes out the bytes stored in the transfer buffer in
 * the board structure using PIO. Also the number of bytes is stored
 * in the board structure as well as the information if this is the
 * last of a series of write operations to the same device. The
 * function expects the device to be addressed as listener and the
 * board as talker and the ATN line to be unasserted.
 *-------------------------------------------------------------------*/

static int board_pio_write( Board_t * board )
{
	unsigned int count = 0;
	unsigned char isr1 = 0;


	PDEBUGin( );

	/* If interrupt can be used the transfer semaphore must be 'armed' and
	   the interrupt must be enabled */

	if ( board->status & USE_IRQ ) {
		semaphore_init( &board->transf.sem, 0 );
		if ( board->type == PCIIA )
			board_out( ( void * ) ( 0x2f0 | board->irq ), 0xff );
		board_out( reg_addr.IMR1, HR_DOIE | HR_ERRIE );
	}

	/* Output all the bytes */

	while ( count < board->transf.cnt && ! board->timeout ) {
		/* We have to set the EOI line with the next byte if this is
		   the very last byte and the eosmode is set to send EOI with
		   the last byte */

		if ( board->transf.is_last && board->transf.cnt - 1 == count &&
		     board->eosmode & GPIB_EOT )
			board_out( reg_addr.AUXMR, AUX_seoi );

		/* Wait for the chip to become ready and send next byte */

		isr1 = board_wait_out( board );
		if ( board->timeout || isr1 & HR_ERR )
			break;
		board_out( reg_addr.CDOR, * ( unsigned char * )
					     ( board->transf.buf + count++ ) );
	}

	/* Wait until the transfer is really finished, i.e. until the DO bit
	   in ISR1 being set again (unless an error or timeout happened) */

	if ( ! ( board->timeout || isr1 & HR_ERR ) )
		isr1 = board_wait_out( board );

	gpib_count = count;

	/* Now disable the interrupt */

	if ( board->status & USE_IRQ )
		board_out( reg_addr.IMR1, 0 );

	/* Set up the status flags */

	if ( board->timeout ) {
		gpib_status |= GPIB_ERR | GPIB_TIMO;
		gpib_error   = GPIB_EABO;
	} else if ( board->transf.is_last && gpib_count == board->transf.cnt )
		gpib_status |= GPIB_CMPL;

	PDEBUG( "board->transf.cnt = %u, count %u\n",
		board->transf.cnt, count );

	PDEBUGout( );
	return gpib_status;
}


/*---------------------------------------------------------------------------*
 * If interrupts can be used simply a down is done on the transfer semaphore
 * - if the interrupt has already come in this does nothing because then the
 * semaphore has been up'ed by the interrupt handler, otherwise the process
 * will sleep until the interrupt handler is triggered. If no interrupts can
 * be used the ISR1 register is polled until CDOR can accept a new byte (as
 * indicated by the HR_DO bit being set) or a timeout occurs.
 *---------------------------------------------------------------------------*/

static unsigned char board_wait_out( Board_t * board )
{
	unsigned char isr1 = 0;


	PDEBUGin( );

	if ( board->status & USE_IRQ ) {
		/* Don't do a down() if a timeout has occurred - the timeout
		   might have actually happend just before the semaphore was
		   initialized! */

		if ( ! board->timeout ) {
			down( &board->transf.sem );
			if ( board->type == PCIIA )
				board_out( ( void * ) ( 0x2f0 | board->irq ),
					   0xff );
		}
		isr1 = board_in( reg_addr.ISR1 );
	} else
		while ( ! ( ( isr1 = board_in( reg_addr.ISR1 ) ) & 
			    ( HR_ERR | HR_DO ) || board->timeout ) )
			/* empty */ ;

	PDEBUGout( );
	return isr1;
}


/*
 * Local variables:
 * c-basic-offset: 8
 * c-indent-level: 8
 * c-brace-imaginary-offset: 0
 * c-brace-offset: 0
 * c-argdecl-indent: 4
 * c-label-ofset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: t
 * tab-width: 8
 * End:
 */
