/*
  $Id: functions.c 737 2007-08-04 14:46:00Z jens $

  Copyright (C) 2001-2006 Jens Thoms Toerring
 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
  To contact the author send email to: jt@toerring.de
*/


#include "gpib_module.h"
#include <linux/fcntl.h>
#include "sys_common.h"
#include "sys.h"
#include "board.h"



int gpib_open( struct inode * inode,
	       struct file *  filep );
#if LINUX_VERSION_CODE < KERNEL_VERSION( 2, 6, 11 )
int gpib_ioctl( struct inode * inode,
		struct file *  filep,
		unsigned int   cmd,
		unsigned long  arg );
#else
long gpib_ioctl( struct file *  filep,
		 unsigned int   cmd,
		 unsigned long  arg );
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 2, 0 )
int gpib_close( struct inode * inode,
		struct file *  filep );
ssize_t gpib_VFS_read( struct file * filep,
		       char __user * buffer,
		       size_t        count,
		       loff_t *      wtf );
ssize_t gpib_VFS_write( struct file *        filep,
			const char __user *  buffer,
			size_t               count,
			loff_t *             wtf );
#else
void gpib_close( struct inode * inode,
		 struct file *  filep );
int gpib_VFS_read( struct inode * inode,
		   struct file *  filep,
		   char *         buffer,
		   int            count );
int gpib_VFS_write( struct inode * inode,
		    struct file *  filep,
		    const char *   buffer,
		    int           count );
#endif

/* Structure file_operations changed betwen the major Linux versions... */

#if LINUX_VERSION_CODE < KERNEL_VERSION( 2, 2, 0 )
struct file_operations gpib_fops = {
	NULL,               /* lseek              */
	gpib_VFS_read,      /* read               */
	gpib_VFS_write,     /* write              */
	NULL,               /* readdir            */
	NULL,               /* select             */
	gpib_ioctl,         /* ioctl              */
	NULL,               /* mmap               */
	gpib_open,          /* open               */
	gpib_close,         /* release            */
	NULL,               /* fsync              */
	NULL,               /* fasync             */
	NULL,               /* check_media_change */
	NULL                /* revalidate         */
};
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 2, 0 ) && \
    LINUX_VERSION_CODE < KERNEL_VERSION( 2, 4, 0 )
struct file_operations gpib_fops = {
	NULL,               /* llseek             */
	gpib_VFS_read,      /* read               */
	gpib_VFS_write,     /* write              */
	NULL,               /* readdir            */
	NULL,               /* poll               */
	gpib_ioctl,         /* ioctl              */
	NULL,               /* mmap               */
	gpib_open,          /* open               */
	NULL,               /* flush              */
	gpib_close,         /* release            */
	NULL,               /* fsync              */
	NULL,               /* fasync             */
	NULL,               /* check_media_change */
	NULL,               /* revalidate         */
	NULL                /* lock               */
};
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 4, 0 ) && \
    LINUX_VERSION_CODE < KERNEL_VERSION( 2, 6, 0 )
struct file_operations gpib_fops = {
	owner:		  THIS_MODULE,
	read:		  gpib_VFS_read,
	write:		  gpib_VFS_write,
	ioctl:		  gpib_ioctl,
	open:		  gpib_open,
	release:	  gpib_close,
};
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 6, 0 )
struct file_operations gpib_fops = {
	.owner =	  THIS_MODULE,
	.read =		  gpib_VFS_read,
	.write =	  gpib_VFS_write,
#if LINUX_VERSION_CODE < KERNEL_VERSION( 2, 6, 11 )
	.ioctl =	  gpib_ioctl,
#else
	.unlocked_ioctl = gpib_ioctl,
#endif
	.open =		  gpib_open,
	.release =	  gpib_close,
};
#endif



static int gpib_lock_board( int nonblock );
static void gpib_unlock_board( void );
static int  get_from_user( unsigned long arg,
			   GPIB_Arg *    gpib_arg );
static void status_reset( void );



/*-------------------------------------------------------------------*
 * Function 'opens' the device file - this function is called either
 * for the board itself or indirectly via the virtual file system
 * calls for one of the devices connected to the bus.
 *-------------------------------------------------------------------*/

int gpib_open( struct inode * inode,
	       struct file *  filep )
{
	int ret;


	PDEBUGin( );

	/* Keep the board from being accessed by more than one process at
	   once - if the board is already in use and thus locked and the
	   device file is opened with either the NONBLOCK or NDELAY flag
	   return EAGAIN, otherwise sleep on a semaphore until the lock
	   is removed by the other process */

	if ( ( ret = gpib_lock_board( filep->f_flags &
				      ( O_NONBLOCK | O_NDELAY ) ) ) < 0 ) {
		PDEBUGout( );
		return ret;
	}

	if ( board.open_count > 0 && filep->f_flags & O_EXCL ) {
		gpib_unlock_board( );
		PDEBUGout( );
		return - EBUSY;
	}

#if LINUX_VERSION_CODE < KERNEL_VERSION( 2, 4, 0 )
	MOD_INC_USE_COUNT;
#endif

	/* Reset the status information */

	status_reset( );

	/* If the board isn't opened for the first time it's already CIC,
	   so return immediately after incrementing the open count. */

	if ( board.open_count++ > 0 ) {
		gpib_unlock_board( );
		PDEBUGout( );
		return 0;
	}

	/* The rest needs to be done only if the module wasn't in use yet */

	board.open_read = board.open_write =
	board.last_read = board.last_write = 0;
	board.last_error = -1;

	/* Get memory for transfers */

	if ( ! gpib_get_memory( &board, mem_size ) ) {
		gpib_status |= GPIB_ERR;
		board.last_error = gpib_error = GPIB_EMEM;
		PDEBUGout( );
		return - ENOMEM;
	}

	/* Make the board controller in charge (CIC) */

	if ( ! ( gpib_status & GPIB_CIC ) && gpib_onl( 1 ) != 0 ) {
		board.last_error = gpib_error;
		board.open_count--;
#if LINUX_VERSION_CODE < KERNEL_VERSION( 2, 4, 0 )
		MOD_DEC_USE_COUNT;
#endif
		gpib_get_memory( &board, 0 );
		gpib_unlock_board( );
		PDEBUGout( );
		return - EIO;
	}

	if ( board.use_irq ) {
		board.status |= USE_IRQ;
		if ( board.use_dma )
			board.status |= USE_DMA;
	}
	else
		board.use_dma = UNSET;

	board.gpib_status = gpib_status;

	/* Let other processes get at the board again */

	gpib_unlock_board( );

	PDEBUGout( );
	return 0;
}


/*----------------------*
 * Closes a device file
 *----------------------*/

#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 2, 0 )
int gpib_close( struct inode * inode,
		struct file *  filep )
#else
void gpib_close( struct inode * inode,
		 struct file *  filep )
#endif
{
	PDEBUGin( );

	/* Keep other processes from interfering while we're busy - in this
	   case we always wait for our turn, even when nonblocking mode was
	   set for the file */

	gpib_lock_board( 0 );

	/* Don't close board if it's not open at all */

	if ( board.open_count == 0 ) {
		gpib_unlock_board( );
		PDEBUGout( );
#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 2, 0 )
		return 0;
#else
		return;
#endif
	}

	/* If there are still other processes having the board open don't
	   bring the board offline but return immediately */

	if ( --board.open_count > 0 )
	{
		status_reset( );

#if LINUX_VERSION_CODE < KERNEL_VERSION( 2, 4, 0 )
		MOD_DEC_USE_COUNT;
#endif
		gpib_unlock_board( );
		PDEBUGout( );
#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 2, 0 )
		return 0;
#else
		return;
#endif
	}

	/* Reset the status information */

	status_reset( );

	/* Bring the board offline */

	if ( gpib_status & GPIB_CIC )
		gpib_onl( 0 );

	gpib_status = gpib_get_state( );
	board.gpib_status = gpib_status;
	if ( gpib_error )
		board.last_error = gpib_error;

#if LINUX_VERSION_CODE < KERNEL_VERSION( 2, 4, 0 )
	MOD_DEC_USE_COUNT;
#endif

	/* Get rid of the memory used for transfers */

	gpib_get_memory( &board, 0 );

	gpib_unlock_board( );

	PDEBUGout( );
#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 2, 0 )
	return 0;
#endif
}


/*--------------------------------------------------------*
 * Send data to a device via the virtual file system call
 *--------------------------------------------------------*/

#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 2, 0 )
ssize_t gpib_VFS_write( struct file *        filep,
			const char __user *  buffer,
			size_t               count,
			loff_t *             wtf )
#else
int gpib_VFS_write( struct inode * inode,
		    struct file *  filep,
		    const char *   buffer,
		    int count )
#endif
{
	unsigned int minor;
	int ret;


	PDEBUGin( );

	/* Keep other processes from interfering while we're busy */

	if ( gpib_lock_board( filep->f_flags & ( O_NONBLOCK | O_NDELAY ) ) ) {
		PDEBUGout( );
		return - EAGAIN;
	}

	/* If board hasn't been opened something strange is going on... */

	if ( board.open_count == 0 ) {
		gpib_unlock_board( );
		PDEBUGout( );
		return - EIO;
	}

	/* Reset the status information */

	status_reset( );

	/* Write out the user supplied buffer */

#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 2, 0 )
	minor = MINOR( filep->f_dentry->d_inode->i_rdev );
#else
	minor = MINOR( inode->i_rdev );
#endif

	ret = gpib_write( minor, buffer, ( unsigned long ) count, OFF );

	/* If an error happened return the error code, otherwise the number
	   of bytes written */

	PDEBUG( "retval = %d, count = %ld\n", ret, gpib_count );

	board.total_write += gpib_count;
	board.open_write += gpib_count;
	board.last_write = gpib_count;
	board.gpib_status = gpib_status;
	if ( gpib_error )
		board.last_error = gpib_error;

	gpib_unlock_board( );

	PDEBUGout( );

#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 2, 0 )
	return ret ? ( ssize_t ) ret : ( ssize_t ) gpib_count;
#else
	return ret ? ret : ( int ) gpib_count;
#endif
}


/*----------------------------------------------------------*
 * Read data from a device via the virtual file system call
 *----------------------------------------------------------*/

#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 2, 0 )
ssize_t gpib_VFS_read( struct file * filep,
		       char __user * buffer,
		       size_t        count,
		       loff_t *      wtf )
#else
int gpib_VFS_read( struct inode * inode,
		   struct file *  filep,
		   char *         buffer,
		   int            count )
#endif
{
	unsigned int minor;
	int ret;


	PDEBUGin( );

	/* Keep other processes from interfering while we're busy */

	if ( gpib_lock_board( filep->f_flags & ( O_NONBLOCK | O_NDELAY ) ) ) {
		PDEBUGout( );
		return - EAGAIN;
	}

	/* If board hasn't been opened something strange is going on... */

	if ( board.open_count == 0 ) {
		gpib_unlock_board( );
		PDEBUGout( );
		return - EIO;
	}

	/* Reset the status information */

	status_reset( );

#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 2, 0 )
	minor = MINOR( filep->f_dentry->d_inode->i_rdev );
#else
	minor = MINOR( inode->i_rdev );
#endif

	ret = gpib_read( minor, buffer, ( unsigned long ) count, OFF );

	board.total_read  += gpib_count;
	board.open_read   += gpib_count;
	board.gpib_status  = gpib_status;
	board.last_read    = gpib_count;

	if ( gpib_error )
		board.last_error = gpib_error;

	PDEBUG( "retval = %d, count = %ld\n", ret, gpib_count );

	/* If an error happened return the error code, otherwise the number
	   of bytes read. On timeout return zero to indicate end of file. */

	if ( gpib_status & GPIB_TIMO ) {
		gpib_unlock_board( );
		PDEBUGout( );
		return 0;
	}

	gpib_unlock_board( );

	PDEBUGout( );

#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 2, 0 )
	return ret != 0 ? ( ssize_t ) ret : ( ssize_t ) gpib_count;
#else
	return ret != 0 ? ret : gpib_count;
#endif
}


/*----------------------------------*/
/* Handler for ioctl() system calls */
/*----------------------------------*/

#if LINUX_VERSION_CODE < KERNEL_VERSION( 2, 6, 11 )
int gpib_ioctl( struct inode * inode,
		struct file *  filep,
		unsigned int   cmd,
		unsigned long  arg )
#else
long gpib_ioctl( struct file *  filep,
		 unsigned int   cmd,
		 unsigned long  arg )
#endif
{
	int ret = 0;                      /* assume everything is OK for now */
	GPIB_Arg gpib_arg;                /* arguments get copied to this */


	PDEBUGin( );

	/* Keep other processes from interfering while we're busy */

	if ( gpib_lock_board( filep->f_flags & ( O_NONBLOCK | O_NDELAY ) ) ) {
		PDEBUGout( );
		return - EAGAIN;
	}

	/* If board hasn't been opened something strange is going on... */

	if ( board.open_count == 0 ) {
		gpib_unlock_board( );
		PDEBUGout( );
		return - EIO;
	}

	/* Reset the status information */

	status_reset( );

	/* Check that we have write access to the parameters passed to the
	   routine (write access because all functions return status or error
	   information via this structure) and copy it from user to kernel
	   space */

	if ( ( ret = get_from_user( arg, &gpib_arg ) ) != 0 ) {
		PDEBUG( "Access to user space data denied\n" );
		gpib_status |= GPIB_ERR;
		gpib_error = GPIB_EDVR;
		gpib_unlock_board( );
		PDEBUGout( );
		return - EFAULT;
	}

	/* Filter out calls with the wrong magic type or undefined commands.
	   As far as I understand SUSV3 what we need to return in these
	   cases is EINVAL. */

	if ( _IOC_TYPE( cmd ) != GPIB_MAGIC_IOC || 
	     _IOC_NR( cmd ) < GPIB_MIN_IOC_NR   ||
	     _IOC_NR( cmd ) > GPIB_MAX_IOC_NR ) {
		PDEBUG( "Invalid ioctl() call\n" );
		gpib_arg.stat = gpib_status | GPIB_ERR; 
		gpib_arg.err  = GPIB_ECAP;

#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 4, 0 )
		copy_to_user( ( void __user * ) arg, &gpib_arg,
			      sizeof gpib_arg );
#elif LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 2, 0 )
		__copy_to_user( ( void * ) arg, &gpib_arg, sizeof gpib_arg );
#else
		memcpy_tofs( ( void * ) arg, &gpib_arg, sizeof gpib_arg ) );
#endif
		gpib_unlock_board( );
		PDEBUGout( );
		return - EINVAL;
	}

	/* Now let's handle the command */

	switch ( cmd ) {
		case GPIB_IOC_READ :
			PDEBUG( "GPIB_IOC_READ\n" );
			ret = gpib_read( gpib_arg.arg,
					 ( void __user * ) gpib_arg.buf,
					 gpib_arg.cnt, OFF );
			board.total_read += gpib_count;
			board.open_read  += gpib_count;
			board.last_read   = gpib_count;
			break;

		case GPIB_IOC_WRT :
			PDEBUG( "GPIB_IOC_WRT\n" );
			ret = gpib_write( gpib_arg.arg,
					  ( void __user * ) gpib_arg.buf,
					  gpib_arg.cnt, OFF );
			board.total_write += gpib_count;
			board.open_write  += gpib_count;
			board.last_write   = gpib_count;
			break;

		case GPIB_IOC_CMD :
			PDEBUG( "GPIB_IOC_CMD\n" );
			ret = gpib_cmd( ( void __user * ) gpib_arg.buf,
					gpib_arg.cnt );
			board.total_write += gpib_count;
			board.open_write  += gpib_count;
			board.last_write   = gpib_count;
			break;

		case GPIB_IOC_DVRD :
			PDEBUG( "GPIB_IOC_DVRD\n" );
			ret = gpib_read( gpib_arg.arg,
					 ( void __user * ) gpib_arg.buf,
					    gpib_arg.cnt, ON );
			board.total_read += gpib_count;
			board.open_read  += gpib_count;
			board.last_read   = gpib_count;
			break;

		case GPIB_IOC_DVWRT :
			PDEBUG( "GPIB_IOC_DVWRT\n" );
			ret = gpib_write( gpib_arg.arg,
					  ( void __user * ) gpib_arg.buf,
					  gpib_arg.cnt, ON );
			board.total_write += gpib_count;
			board.open_write  += gpib_count;
			board.last_write   = gpib_count;
			break;

		case GPIB_IOC_STATE :
			PDEBUG( "GPIB_IOC_STATE\n" );
			gpib_get_state( );
			break;

		case GPIB_IOC_WAIT :
			ret = gpib_wait( gpib_arg.arg );
			break;

		case GPIB_IOC_PPC :
			PDEBUG( "GPIB_IOC_PPC\n" );
			ret = gpib_ppc( gpib_arg.arg, gpib_arg.stat );
			break;

		case GPIB_IOC_RPP :
			PDEBUG( "GPIB_IOC_RPP\n" );
			ret = gpib_rpp( ( void __user * ) gpib_arg.buf );
			break;

		case GPIB_IOC_ONL :                 /* bring board online */
			PDEBUG( "GPIB_IOC_ONL\n" );
			ret = gpib_onl( gpib_arg.arg );
			break;

		case GPIB_IOC_SIC :                 /* raise the IFC line */
			PDEBUG( "GPIB_IOC_SIC\n" );
			ret = gpib_sic( );
			break;

		case GPIB_IOC_REN :                 /* set or reset REN line */
			PDEBUG( "GPIB_IOC_REN\n" );
			ret = gpib_ren( gpib_arg.arg != 0 ? SET : UNSET );
			break;

		case GPIB_IOC_GTS :                 /* release the ATN line */
			PDEBUG( "GPIB_IOC_GTS\n" );
			ret = gpib_gts( );
			break;

		case GPIB_IOC_CAC :
			PDEBUG( "GPIB_IOC_CAC\n" );
			ret = gpib_cac( gpib_arg.arg );
			break;

		case GPIB_IOC_LINES :
			PDEBUG( "GPIB_IOC_LINES\n" );
			gpib_arg.arg = gpib_lines( );
			break;

		case GPIB_IOC_PAD :
			PDEBUG( "GPIB_IOC_PAD\n" );
			ret = gpib_pad( gpib_arg.arg );
			break;

		case GPIB_IOC_SAD :
			PDEBUG( "GPIB_IOC_SAD\n" );
			ret = gpib_sad( gpib_arg.arg );
			break;

		case GPIB_IOC_TIMO :
			PDEBUG( "GPIB_IOC_TIMO\n" );
			ret = gpib_timo( gpib_arg.arg );
			break;

		case GPIB_IOC_DMA :
			PDEBUG( "GPIB_IOC_DMA\n" );
			if ( gpib_arg.arg >= 0 )
				ret = gpib_set_dma( gpib_arg.arg );
			else
				gpib_arg.arg =
					    ( board.status & USE_DMA ) ? 1 : 0;
			break;

		case GPIB_IOC_IRQ :
			PDEBUG( "GPIB_IOC_IRQ\n" );
			if ( gpib_arg.arg >= 0 )
				ret = gpib_set_irq( gpib_arg.arg );
			else
				gpib_arg.arg =
					    ( board.status & USE_IRQ ) ? 1 : 0;
			break;

		case GPIB_IOC_EOS :
			PDEBUG( "GPIB_IOC_EOS\n" );
			ret = gpib_eos( gpib_arg.arg );
			break;

		case GPIB_IOC_TRG :         /* Send trigger to a device */
			PDEBUG( "GPIB_IOC_TRG\n" );
			ret = gpib_trg( ( unsigned int ) gpib_arg.arg );
			break;

		case GPIB_IOC_SDC :         /* Send clear to a device */
			PDEBUG( "GPIB_IOC_SDC\n" );
			ret = gpib_clr( ( unsigned int ) gpib_arg.arg );
			break;

		case GPIB_IOC_DCL :          /* Send clear to all devices */
			PDEBUG( "GPIB_IOC_DCL\n" );
			ret = gpib_dcl( );
			break;

		case GPIB_IOC_LOCAL :        /* Unlock a device */
			PDEBUG( "GPIB_IOC_LOC\n" );
			ret = gpib_local( ( unsigned int ) gpib_arg.arg );
			break;

		case GPIB_IOC_LLO :          /* Send local lockout to device */
			PDEBUG( "GPIB_IOC_LLO\n" );
			ret = gpib_llo( ( unsigned int ) gpib_arg.arg );
			break;

		case GPIB_IOC_RSP :
			PDEBUG( "GPIB_IOC_RSP\n" );
			ret = gpib_rsp( ( unsigned int ) gpib_arg.arg );
			gpib_arg.arg = gpib_count;
			break;

		case GPIB_IOC_LN :           /* find listener */
			PDEBUG( "GPIB_IOC_LN\n" );
			gpib_arg.arg =
				      gpib_ln( ( unsigned int ) gpib_arg.arg );
			break;

		default :                  /* we should never end up here... */
			PDEBUG( "Invalid ioctl() call\n" );
			ret = - EINVAL;
	}

	board.gpib_status = gpib_status;
	if ( gpib_error )
		board.last_error = gpib_error;

	gpib_arg.stat = gpib_status; 
	gpib_arg.err  = gpib_error;
	gpib_arg.cnt  = gpib_count;

#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 4, 0 )
	__copy_to_user( ( void __user * ) arg, &gpib_arg, sizeof gpib_arg );
#elif LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 2, 0 )
	__copy_to_user( ( void * ) arg, &gpib_arg, sizeof gpib_arg );
#else
	memcpy_tofs( ( void * ) arg, &gpib_arg, sizeof gpib_arg );
#endif

	gpib_unlock_board( );

	PDEBUGout( );
	return ret;
}


/*------------------------------------------------------------------------*
 * This function is used to keep more than one process from accessing the
 * board. So whenever the board gets accessed a semaphore is decremented.
 * If the function is called with 'nonblock' being set we actually don't
 * do an unconditioned down on the semaphore but instead return immedia-
 * tely if the down would put the process to sleep with a return value of
 * 1 to signal the problem. (Unfortunately, this does not work for older
 * kernels which didn't had a down_trylock() function. For these kernel
 * versions the nonblock flag is simply ignored, i.e. with these kernels
 * the function trying to access the board will always sleep until the
 * other process holding the lock is done.)
 *------------------------------------------------------------------------*/

static int gpib_lock_board( int nonblock )
{
	int ret = 0;


	PDEBUGin( );

#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 2, 0 )
	if ( nonblock && down_trylock( & board.lock_sem ) )
		ret = - EAGAIN;
	else if ( ! nonblock && down_interruptible( &board.lock_sem ) )
		ret = - ERESTARTSYS;
#else
	down( &board.lock_sem );
#endif
	     
	PDEBUGout( );
	return ret;
}


/*-----------------------------------------------------*/
/* Called by a process when it's finished to let other */
/* processes continue with whatever they have to do.   */
/*-----------------------------------------------------*/

static void gpib_unlock_board( void )
{
	PDEBUGin( );
	up( &board.lock_sem );
	PDEBUGout( );
}


/*--------------------------------------------------------------------*
 * Copy the argument passed to the ioctl() call to kernel space while
 * checking write access to it - this function does just this in the
 * way appropriate to the kernel version.
 *--------------------------------------------------------------------*/

static int get_from_user( unsigned long arg,
			  GPIB_Arg *    gpib_arg )
{
	int ret = 0;


#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 4, 0 )
        if ( copy_from_user( gpib_arg, ( const void __user * ) arg,
                             sizeof *gpib_arg ) )
                ret =  -EFAULT;
#elif LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 2, 0 )
	if ( access_ok( VERIFY_READ, ( void * ) arg, sizeof *gpib_arg ) )
		 __copy_from_user( gpib_arg, ( void * ) arg,
				   sizeof *gpib_arg ) );
	else
		ret = -EFAULT;
#else
	if ( ( ret = verify_area( VERIFY_WRITE, ( void * ) arg,
				  sizeof *gpib_arg) ) == 0 ) 
		memcpy_fromfs( gpib_arg, ( void * ) arg, sizeof *gpib_arg );
#endif

	return ret;
}


/*-------------------------------------------------------------*
 *-------------------------------------------------------------*/

static void status_reset( void )
{
	gpib_status &= ~ ( GPIB_TIMO | GPIB_CMPL | GPIB_ERR | GPIB_END );
	gpib_error   = 0;
	gpib_count   = 0;

	board.gpib_status = gpib_status;
	board.timeout     = UNSET;
	board.gpib_status = gpib_status;
}


/*
 * Overrides for Emacs so that we follow the Linux tabbing style
 *
 * Local variables:
 * c-basic-offset: 8
 * c-indent-level: 8
 * c-brace-imaginary-offset: 0
 * c-brace-offset: 0
 * c-argdecl-indent: 4
 * c-label-ofset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: t
 * tab-width: 8
 * End:
 */
