/*
  $Id: read.c 732 2006-05-30 20:13:39Z jens $

  Copyright (C) 2001-2006 Jens Thoms Toerring
 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
  To contact the author send email to: jt@toerring.de
*/


#include "gpib_module.h"
#include "sys_common.h"
#include "sys.h"
#include "board.h"



/*----------------------------------------------------------------------*
 * Function reads a number of bytes from a device on the GPIB bus. The
 * address of the device, the buffer for storing the bytes read and the
 * maximum number of bytes are the first three arguments. If the last
 * argument is set, the device has already to be is addressed and the
 * board has to be in CIC state.
 *-----------------------------------------------------------------------*/

int gpib_read( unsigned int  address,
	       void __user * buffer,
	       unsigned long read_count,
	       bool          dvrd )
{
	void __user *buf;
	unsigned long to_be_read,
		      already_read;
	unsigned int chunk_size;


	PDEBUGin( );

	/* Just return if there's actually nothing to be read */

	if ( read_count == 0 ) {
		gpib_status |= GPIB_CMPL;
		gpib_count = 0;
		PDEBUGout( );
		return 0;
	}

	/* We can't do anything if the board isn't controller */

	if ( ! ( board_get_state( &board ) & GPIB_CIC ) ) {
		gpib_error = GPIB_ECIC;
		PDEBUG( "Board not CIC\n" );
		PDEBUGout( );
		return - EIO;
	}

	/* Check if we have write access to the user supplied buffer */

#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 2, 0 )
	if ( access_ok( VERIFY_WRITE, buffer, read_count ) == 0 ) {
#else
	if ( verify_area( VERIFY_WRITE, buffer, read_count ) != 0 ) {
#endif
		gpib_status |= GPIB_ERR;
		gpib_error   = GPIB_EARG;
		PDEBUGout( );
		return - EFAULT;
	}

	/* If the device read flag is unset set device to be talker and board
	   to be listener, otherwise this should have already been done */

	if ( ! dvrd && gpib_setup_talker( address ) & GPIB_ERR ) {
		PDEBUGout( );
		return - EIO;
	}
	else if ( dvrd ) {
		if ( ! ( gpib_status & GPIB_LACS ) ) {
			PDEBUG( "Board not listener\n" );
			gpib_status |= GPIB_ERR;
			gpib_error   = GPIB_EADR;
			PDEBUGout( );
			return - EIO;
		}

		/* Release ATN line if it's still asserted */

		if ( gpib_status & GPIB_ATN &&
		     board_gts( &board ) & GPIB_ERR ) {
			PDEBUGout( );
			return - EIO;
		}
	}

	/* If the transfer buffer isn't large enough to hold all the data to
	   be read we have to split the transfer into chunks of the buffers
	   size */

	gpib_count = 0;
	to_be_read = read_count;
	already_read = 0;
	buf = buffer;
	if ( to_be_read > board.buf_size ) {
		chunk_size = board.buf_size;
		board.transf.is_last = UNSET;
	} else {
		chunk_size = to_be_read;
		board.transf.is_last = SET;
	}

	/* Set up the fields of the transfer structure that won't change */

	board.transf.buf = board.buf;
	board.transf.dir = DMA_RD;

	/* Start the timer to avoid 'hanging' due to unresponsive devices */

	gpib_start_timer( );

	/* And, finally, here we go: Read data from the bus, copy them to the
	   user buffer and repeat until either all data are written or an error
	   occurred */

	do {
		board.transf.cnt = chunk_size;

		/* Now read the data */

		board_read( &board );

		/* Copy data to user space */

#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 2, 0 )
		__copy_to_user( buf, board.transf.buf, gpib_count );
#else
		memcpy_tofs( buf, board.transf.buf, gpib_count );
#endif

		/* Calculate how much still needs to be read */

		already_read += gpib_count;
		to_be_read   -= gpib_count;

		/* Increment the pointer to user space to where the remaining
		   data have to be written to */

		buf += gpib_count;
		gpib_count = 0;

		if ( to_be_read <= chunk_size ) {
			chunk_size = to_be_read;
			board.transf.is_last = SET;
		}

	} while ( to_be_read > 0 &&
		  ! ( gpib_status & ( GPIB_ERR | GPIB_END ) ) );

	/* Stop the timer and store number of bytes read in in gpib_count */

	gpib_stop_timer( );
	gpib_count = already_read;
	if ( ! ( gpib_status & GPIB_ERR ) )
		gpib_status |= GPIB_CMPL;
	else
		gpib_status &= ~GPIB_CMPL;

	PDEBUG( "Read %lu, returning %d, gpib_status = 0x%x\n", gpib_count,
		( gpib_status & GPIB_ERR ) ? - EIO : 0, gpib_status );
	PDEBUGout( );
	return ( gpib_status & GPIB_ERR ) ? - EIO : 0;
}


/*
 * Local variables:
 * c-basic-offset: 8
 * c-indent-level: 8
 * c-brace-imaginary-offset: 0
 * c-brace-offset: 0
 * c-argdecl-indent: 4
 * c-label-ofset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: t
 * tab-width: 8
 * End:
 */
