/*
  $Id: timer.c 716 2006-03-07 10:27:35Z jens $

  Copyright (C) 2001-2006 Jens Thoms Toerring
 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
  To contact the author send email to: jt@toerring.de
*/

#include "gpib_module.h"
#include "sys_common.h"
#include "sys.h"
#include "board.h"


static int time_table[ 18 ];

static void gpib_timer_handler( unsigned long unused );


/*----------------------------------------------------------------------*
 * Function for setting up the table for the different timer settings.
 * Actually, on most systems the timer is checked only 100 times a
 * second (or whatever the systems basic rate is), so all settings
 * below a jiffy (on i386 about 10 ms) are more or less useless and
 * will wait for a jiffy (that's the reason why 1 is added, i.e. to
 * make the timer wait at least that long).
 *----------------------------------------------------------------------*/

void gpib_setup_timer( void )
{
	board.timeout = UNSET;

	time_table[  1 ] = ( HZ *   10 ) / 1000000L + 1;    /*  1: T10us    */
	time_table[  2 ] = ( HZ *   30 ) / 1000000L + 1;    /*  2: T30us    */
	time_table[  3 ] = ( HZ *  100 ) / 1000000L + 1;    /*  3: T100us   */
	time_table[  4 ] = ( HZ *  300 ) / 1000000L + 1;    /*  4: T300us   */
	time_table[  5 ] = ( HZ *    1 ) / 1000 + 1;        /*  5: T1ms     */
	time_table[  6 ] = ( HZ *    3 ) / 1000 + 1;        /*  6: T3ms     */
	time_table[  7 ] = ( HZ *   10 ) / 1000 + 1;        /*  7: T10ms    */
	time_table[  8 ] = ( HZ *   30 ) / 1000 + 1;        /*  8: T30ms    */
	time_table[  9 ] = ( HZ *  100 ) / 1000 + 1;        /*  9: T100ms   */
	time_table[ 10 ] = ( HZ *  300 ) / 1000 + 1;        /* 10: T300ms   */
	time_table[ 11 ] = ( HZ *    1 ) + 1;               /* 11: T1s      */
	time_table[ 12 ] = ( HZ *    3 ) + 1;               /* 12: T3s      */
	time_table[ 13 ] = ( HZ *   10 ) + 1;               /* 13: T10s     */
	time_table[ 14 ] = ( HZ *   30 ) + 1;               /* 14: T30s     */
	time_table[ 15 ] = ( HZ *  100 ) + 1;               /* 15: T100s    */
	time_table[ 16 ] = ( HZ *  300 ) + 1;               /* 16: T300s    */
	time_table[ 17 ] = ( HZ * 1000 ) + 1;               /* 17: T1000s   */
}


/*-------------------------------------------*
 * Function for setting a new timer duration
 *-------------------------------------------*/

int gpib_timo( int timo )
{
	PDEBUGin( );

	if ( timo < TNONE || timo > T1000s ) {
		gpib_status |= GPIB_ERR;
		gpib_error   = GPIB_EARG;
		return - EINVAL;
	}

	board.timer_index = timo;

	PDEBUGout( );
	return 0;
}


/*---------------------------------------------------------------*
 * Function for starting the timer (if it's not already running)
 *---------------------------------------------------------------*/

void gpib_start_timer( void )
{
	PDEBUGin( );

	/* Reset the timeout flag */

	board.timeout = UNSET;

	/* If the timeout is set to infinite or the timer is already running
	   nothing has to be done */

	if ( board.status & TIMER_ON ) {
		PDEBUG( "Timer already running\n" );
		PDEBUGout( );
		return;
	}

	if ( board.timer_index == 0 ) {
		PDEBUG( "Timer set to infinite\n" );
		PDEBUGout( );
		return;
	}

	/* Set the timeout period and start the timer */

	init_timer( &board.timer );
	board.timer.function = gpib_timer_handler;
	board.timer.expires  = jiffies + time_table[ board.timer_index ];
	add_timer( &board.timer );
	board.status |= TIMER_ON;

	PDEBUG( "Timer started with %d\n", board.timer_index );
	PDEBUGout( );
}


/*-----------------------------------------*
 * Function for stopping the running timer
 *-----------------------------------------*/

void gpib_stop_timer( void )
{
	PDEBUGin( );

	/* Remove the timer (if it's running) */

	if ( board.status & TIMER_ON ) {
		if ( ! board.timeout )
			del_timer( &board.timer );
		board.status &= ~ TIMER_ON;
	}

	PDEBUGout( );
}


/*-------------------------------------------------------------------*
 * Function to handle timer events - if interrupts are used up() is
 * called on the transfer semaphore, otherwise only a flag gets set,
 * indicating that the timer kicked in.
 *-------------------------------------------------------------------*/

static void gpib_timer_handler( unsigned long unused )
{
	unused = unused;      /* keeps the compiler happy... */


	PDEBUGin( );

	/* Do an up on the transfer semaphore if IRQ is used - while transfers
	   are underway the semaphore is down()'ed and the program waits for
	   either the transfer ending (which, in turn, triggers an interrupt,
	   'up()'ing the semaphore) or a timeout that brought us here */

	PDEBUG( "TIMER kicked in!\n" );

	if ( board.status & USE_IRQ )
        	up( & ( board.transf.sem ) );

	board.timeout = SET;

	PDEBUGout( );
}


/*
 * Local variables:
 * c-basic-offset: 8
 * c-indent-level: 8
 * c-brace-imaginary-offset: 0
 * c-brace-offset: 0
 * c-argdecl-indent: 4
 * c-label-ofset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: t
 * tab-width: 8
 * End:
 */
