/*
  $Id: write.c 732 2006-05-30 20:13:39Z jens $

  Copyright (C) 2001-2006 Jens Thoms Toerring
 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
  To contact the author send email to: jt@toerring.de
*/


#include "gpib_module.h"
#include "sys_common.h"
#include "sys.h"
#include "board.h"



/*-----------------------------------------------------------------------*
 * Function writes a number of bytes to a device on the GPIB bus. The
 * address of the device, the buffer with the bytes to be written and
 * the number of bytes are the first three arguments. If the last
 * argument is set, the device and board have already to be is addressed
 * as listener and talker.
 *-----------------------------------------------------------------------*/

int gpib_write( unsigned int         address,
		const void __user * buffer,
		unsigned long       write_count,
		bool                dvwrt )
{
	unsigned long to_be_written,
		      already_written;
	unsigned int chunk_size;
	const unsigned char __user *buf;


	PDEBUGin( );

	/* Just return if nothing's there to be written */

	if ( write_count == 0 ) {
		gpib_status |= GPIB_CMPL;
		gpib_count = 0;
		PDEBUGout( );
		return 0;
	}

	/* Can't write if the board isn't controller */

	if ( ! ( board_get_state( &board ) & GPIB_CIC ) ) {
		gpib_error = GPIB_ECIC;
		PDEBUG( "Board not CIC\n" );
		PDEBUGout( );
		return - EIO;
	}

	/* Check if we have read access to the user supplied buffer */

#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 2, 0 )
	if ( access_ok( VERIFY_READ, buffer, write_count ) == 0 ) {
#else
	if ( verify_area( VERIFY_READ, buffer, write_count ) != 0 ) {
#endif
		gpib_status |= GPIB_ERR;
		gpib_error   = GPIB_EARG;
		PDEBUGout( );
		return - EFAULT;
	}

	/* If the device write flag isn't set set device to be listener and
	   board to be talker, otherwise this has already been done */

	if ( ! dvwrt && gpib_setup_listener( address ) & GPIB_ERR ) {
		PDEBUGout( );
		return - EIO;
	} else if ( dvwrt ) {
		if ( ! ( gpib_status & GPIB_TACS ) ) {
			PDEBUG( "Board not talker\n" );		
			gpib_status |= GPIB_ERR;
			gpib_error   = GPIB_EADR;
			PDEBUGout( );
			return - EIO;
		}

		/* Release ATN line if it's still asserted */

		if ( gpib_status & GPIB_ATN &&
		     board_gts( &board ) & GPIB_ERR ) {
			PDEBUGout( );
			return - EIO;
		}
	}

	/* If the buffer isn't large enough to hold all the data to be written
	   split transfer into chunks of the buffers size */

	gpib_count = 0;
	to_be_written = write_count;
	already_written = 0;
	buf = buffer;
	if ( to_be_written > board.buf_size ) {
		chunk_size = board.buf_size;
		board.transf.is_last = UNSET;
	} else {
		chunk_size = to_be_written;
		board.transf.is_last = SET;
	}

	/* Set up the parts of the transfer structure that won't change */

	board.transf.buf = board.buf;
	board.transf.dir = DMA_WR;

	/* Start the timer to avoid hanging due to an unresponsive device */

	gpib_start_timer( );

	/* Now copy data from the user buffer, write them onto the bus and
	   repeat until either all data are written or an error happened */

	do {

#if LINUX_VERSION_CODE >= KERNEL_VERSION( 2, 2, 0 )
		__copy_from_user( board.transf.buf, buf, chunk_size );
#else
		memcpy_fromfs( board.transf.buf, buf, chunk_size );
#endif

		board.transf.cnt = chunk_size;

		/* Write out the data */

		board_write( &board );

		/* Calculate how much still needs to be written */

		already_written += gpib_count;
		to_be_written   -= gpib_count;
			
		/* Increment the pointer to user space to where the remaining
		   data are located */

		buf += gpib_count;
		gpib_count = 0;

		if ( to_be_written <= chunk_size ) {
			chunk_size = to_be_written;
			board.transf.is_last = SET;
		}

	} while ( to_be_written > 0 && ! ( gpib_status & GPIB_ERR ) );

	/* Stop the timer and store number of written bytes in gpib_count */

	gpib_stop_timer( );
	gpib_count = already_written;

	PDEBUG( "Returning %d, gpib_status = 0x%x\n",
		( gpib_status & GPIB_ERR ) ? - EIO : 0, gpib_status );
	PDEBUGout( );
	return ( gpib_status & GPIB_ERR ) ? - EIO : 0;
}


/*
 * Local variables:
 * c-basic-offset: 8
 * c-indent-level: 8
 * c-brace-imaginary-offset: 0
 * c-brace-offset: 0
 * c-argdecl-indent: 4
 * c-label-ofset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: t
 * tab-width: 8
 * End:
 */
