/*
  $Id: device.c 736 2006-07-07 15:48:51Z jens $

  Copyright (C) 2001-2006 Jens Thoms Toerring
 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
  To contact the author send email to: jt@toerring.de
*/


#include "lib.h"


extern Device devices[ ];
extern Device devices_orig[ GPIB_MAX_DEVICES ];
extern int master_file;
extern int num_devices;
extern int master_dev;

extern int gpib_status;
extern int gpib_error;
extern unsigned long gpib_count;

extern bool board_is_online;


int gpib_pad( int dev_dscr,
              int pad );
int gpib_sad( int dev_dscr,
              int sad );
int gpib_loc( int dev_dscr );
int gpib_llo( int dev_dscr );
int gpib_trg( int dev_dscr );
int gpib_clr( int dev_dscr );
int gpib_rsp( int             dev_dscr,
              unsigned char * c );
int gpib_ask( int   dev_dscr,
              int   option,
              int * value );
int gpib_onl( int dev_dscr,
              int v );
int gpib_ppc( int dev_dscr,
              int v );

extern int gpib_ioctl( int cmd,
                       ... );
extern Device * check_dscr( int dev_dscr );
extern int gpib_setup_timo( Device * dev );



/*---------------------------------------*
 * Sets the primary address of a device.
 *---------------------------------------*/

int gpib_pad( int dev_dscr,
              int pad )
{
        Device *dev;


        if ( ( dev = check_dscr( dev_dscr ) ) == NULL )
                return gpib_status;

        /* Check that the address has a reasonable value */

        if ( pad < 0 || pad > 0x1e ) {
                gpib_status |= GPIB_ERR;
                gpib_error   = GPIB_EARG;
                return gpib_status;
        }

        /* If this is the master device we have to really change the boards
           settings (i.e. change the settings of the chips address register),
           while for 'normal' devices only the corresponding entry in the
           device structure has to be changed */

        if ( dev->flags & IS_MASTER &&
             gpib_ioctl( GPIB_IOC_PAD, pad ) & GPIB_ERR )
                return gpib_status;

        dev->pad = pad;

        return gpib_status;
}


/*-----------------------------------------*
 * Sets the secondary address of a device.
 *-----------------------------------------*/

int gpib_sad( int dev_dscr,
              int sad )
{
        Device *dev;


        if ( ( dev = check_dscr( dev_dscr ) ) == NULL )
                return gpib_status;

        /* Check that the address has a reasonable value */

        if ( sad != NO_SAD && (  sad < 0x60 || sad > 0x7e ) ) {
                gpib_status |= GPIB_ERR;
                gpib_error   = GPIB_EARG;
                return gpib_status;
        }

        /* If this is the master device we have to really change the boards
           settings (i.e. change the settings of the chips address register
           and, possibly, the addressing mode), while for 'normal' devices only
           the corresponding entry in the device structure has to be changed */

        if ( dev->flags & IS_MASTER &&
             gpib_ioctl( GPIB_IOC_PAD, sad ) & GPIB_ERR )
                return gpib_status;

        dev->sad = sad;

        return gpib_status;
}


/*------------------------------------------------------------*
 * Brings a device in the local state. Take care: the device
 * goes back into remote state the next time it is addressed.
 *------------------------------------------------------------*/

int gpib_loc( int dev_dscr )
{
        Device *dev;


        if ( ( dev = check_dscr( dev_dscr ) ) == NULL )
                return gpib_status;

        /* The controller can't be brought into the local state */

        if ( dev->flags & IS_MASTER )
        {
                gpib_status |= GPIB_ERR;
                gpib_error   = GPIB_EARG;
                return gpib_status;
        }

        /* Set timeout value */

        if ( gpib_setup_timo( dev ) & GPIB_ERR )
                return gpib_status;

        return gpib_ioctl( GPIB_IOC_LOCAL, ADDRESS( dev->pad, dev->sad ) );
}


/*---------------------------------------------*
 * Brings a device in the local lockout state.
 *---------------------------------------------*/

int gpib_llo( int dev_dscr )
{
        Device *dev;


        if ( ( dev = check_dscr( dev_dscr ) ) == NULL )
                return gpib_status;

        /* The controller can't be brought into the local lockout state */

        if ( dev->flags & IS_MASTER )
        {
                gpib_status |= GPIB_ERR;
                gpib_error   = GPIB_EARG;
                return gpib_status;
        }

        /* Set timeout value */

        if ( gpib_setup_timo( dev ) & GPIB_ERR )
                return gpib_status;

        return gpib_ioctl( GPIB_IOC_LLO, ADDRESS( dev->pad, dev->sad ) );
}


/*--------------------------------------*
 * Sends device a trigger signal (GET).
 *--------------------------------------*/

int gpib_trg( int dev_dscr )
{
        Device *dev;


        if ( ( dev = check_dscr( dev_dscr ) ) == NULL )
                return gpib_status;

        /* Can't send GET to controller */

        if ( dev->flags & IS_MASTER )
        {
                gpib_status |= GPIB_ERR;
                gpib_error   = GPIB_EARG;
                return gpib_status;
        }

        /* Set timeout value set for the master device */

        if ( gpib_setup_timo( dev ) & GPIB_ERR )
                return gpib_status;

        return gpib_ioctl( GPIB_IOC_TRG, ADDRESS( dev->pad, dev->sad ) );
}


/*------------------------------------*
 * Sends device a clear dignal (SDC).
 *------------------------------------*/

int gpib_clr( int dev_dscr )
{
        Device *dev;


        if ( ( dev = check_dscr( dev_dscr ) ) == NULL )
                return gpib_status;

        /* Can't send SDC to controller */

        if ( dev->flags & IS_MASTER )
        {
                gpib_status |= GPIB_ERR;
                gpib_error   = GPIB_EARG;
                return gpib_status;
        }

        /* Set timeout value */

        if ( gpib_setup_timo( dev ) & GPIB_ERR )
                return gpib_status;

        return gpib_ioctl( GPIB_IOC_SDC, ADDRESS( dev->pad, dev->sad ) );
}


/*-----------------------*
 * Serial polls a device
 *-----------------------*/

int gpib_rsp( int            dev_dscr,
              unsigned char * c )
{
        Device *dev;
        int retval;


        if ( ( dev = check_dscr( dev_dscr ) ) == NULL )
                return gpib_status;

        /* Can't serial poll the controller */

        if ( dev->flags & IS_MASTER )
        {
                gpib_status |= GPIB_ERR;
                gpib_error   = GPIB_EARG;
                return gpib_status;
        }

        /* Set timeout for device */

        if ( gpib_setup_timo( dev ) & GPIB_ERR )
                return gpib_status;

        /* Ask for the status byte via serial poll */

        retval = gpib_ioctl( GPIB_IOC_RSP, ADDRESS( dev->pad, dev->sad ) );

        if ( ! ( gpib_status & GPIB_ERR ) )
                *c = ( unsigned char ) ( retval & 0xFF );

        return gpib_status;
}


/*---------------------------------------------------------------*
 * Function for requesting informations about the current device
 * or GPIB state.
 *---------------------------------------------------------------*/

int gpib_ask( int dev_dscr, int option, int *value )
{
        Device *dev;


        /* Get current value of gpib_state and complain if the board isn't
           online. */

        if ( board_is_online )
                gpib_ioctl( GPIB_IOC_STATE );
        else {
                gpib_status &= ~ ( GPIB_CMPL | GPIB_END | GPIB_TIMO );
                gpib_status |= GPIB_ERR;
                gpib_error   = GPIB_EDVR;
                return gpib_status;
        }

        /* Check that the device descriptor is ok */

        if ( dev_dscr < 0 || dev_dscr >= num_devices ) {
                gpib_status |= GPIB_ERR;
                gpib_error   = GPIB_EARG;
                return gpib_status;
        }

        dev = &devices[ dev_dscr ];

        switch ( option ) {
                case GPIB_ASK_SAD :
                        *value = ( int ) dev->sad;
                        break;

                case GPIB_ASK_PAD :
                        *value = ( int ) dev->pad;
                        break;

                case GPIB_ASK_EOS :
                        *value = ( int ) dev->eos;
                        break;

                case GPIB_ASK_XEOS :
                        *value = ( dev->eosmode & GPIB_XEOS ) ? 1 : 0;
                        break;

                case GPIB_ASK_REOS :
                        *value = ( dev->eosmode & GPIB_REOS ) ? 1 : 0;
                        break;

                case GPIB_ASK_BIN :
                        *value = ( dev->eosmode & GPIB_BIN ) ? 1 : 0;
                        break;

                case GPIB_ASK_EOT :
                        *value = ( dev->eosmode & GPIB_EOT ) ? 1 : 0;
                        break;

                case GPIB_ASK_TIMO :
                        *value = dev->timo;
                        break;
                        
                case GPIB_ASK_IS_INIT :
                        *value = ( dev->flags & IS_INIT ) ? 1 : 0;
                        break;

                case GPIB_ASK_IS_MASTER :
                        *value = ( dev->flags & IS_MASTER ) ? 1 : 0;
                        break;

                case GPIB_ASK_IRQ :
                        *value = gpib_ioctl( GPIB_IOC_IRQ, -1 );
                        break;

                case GPIB_ASK_DMA :
                        *value = gpib_ioctl( GPIB_IOC_DMA, -1 );
                        break;
                default :
                        gpib_error   = GPIB_EARG;
                        gpib_status |= GPIB_ERR;
        }

        return gpib_status;
}


/*-----------------------------------------------------------------------*
 * Contrary to what one might guess from its name this function does not
 * bring the board online (or offline). Instead it resets the data for
 * the device or board to the original state (i.e. the one read from the
 * configuration file. If called with 'v' set to zero the device will be
 * brought in the local state, if called for the master device the board
 * will be brought offline, i.e the device file for board accesses is
 * closed.
 *-----------------------------------------------------------------------*/

int gpib_onl( int dev_dscr,
              int v )
{
        Device *dev;


        if ( ( dev = check_dscr( dev_dscr ) ) == NULL )
                return gpib_status;

        if ( dev_dscr != master_dev ) {                   /* normal devices */
                if ( v == 0 && dev->flags & IS_INIT ) {
                        gpib_ioctl( GPIB_IOC_LOCAL,
                                    ADDRESS( dev->pad, dev->sad ) );
                        dev->flags &= ~ IS_INIT;
                }

                if ( v != 0 )
                        memcpy( dev, &devices_orig[ dev_dscr ],
                                sizeof( Device ) );
        } else {                                          /* controller */
        
                if ( v != 0 ) {
                        memcpy( dev, &devices_orig[ dev_dscr ],
                                sizeof( Device ) );
                        dev->flags |= IS_INIT;
                } else {
                        close( master_file );
                        master_dev = -1;
                        num_devices = 0;
                        board_is_online = UNSET;
                        gpib_status = GPIB_LACS;
                }
        }

        return gpib_status;
}


/*-----------------------------------------------------------------------*
 *-----------------------------------------------------------------------*/

int gpib_ppc( int dev_dscr,
              int v )
{
        Device *dev;


        if ( ( dev = check_dscr( dev_dscr ) ) == NULL )
                return gpib_status;

        /* Can't parallel configure the controller */

        if ( dev->flags & IS_MASTER )
        {
                gpib_status |= GPIB_ERR;
                gpib_error   = GPIB_EARG;
                return gpib_status;
        }

        /* Second argument must be a valid parallel poll enable/disable
           command */

        if ( ( v & ~ 0x0F ) != PPE && ( v & ~ 0x0F ) != PPD )
        {
                gpib_status |= GPIB_ERR;
                gpib_error   = GPIB_EARG;
                return gpib_status;
        }

        gpib_ioctl( GPIB_IOC_PPC, ADDRESS( dev->pad, dev->sad ) );

        return gpib_status;
}


/*
 * Local variables:
 * c-basic-offset: 8
 * c-indent-level: 8
 * c-brace-imaginary-offset: 0
 * c-brace-offset: 0
 * c-argdecl-indent: 4
 * c-label-ofset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: nil
 * tab-width: 8
 * End:
 */
