/*
  $Id: lib.h 662 2005-07-26 23:16:49Z jens $

  Copyright (C) 2001-2005 Jens Thoms Toerring
 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
  To contact the author send email to: jt@toerring.de
*/


#define _GNU_SOURCE 1

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <limits.h>
#include <string.h>
#include <sys/ioctl.h>

#include "gpib.h"
#include "autoconf.h"


#define GPIB_MAX_DEVICES 256

#define GPIB_DEF_CONFIG_FILE "/etc/gpib.conf"
#define GPIB_DEF_DEVICE_DIR  "/dev/gpib0"


#define IS_INIT     ( 1 << 0 )
#define IS_MASTER   ( 1 << 1 )

#define ADDRESS( pad, sad ) \
        ( ( ( pad ) & 0xff ) | ( ( ( sad ) & 0xff ) << 8 ) )


typedef enum
{
	false = 0,
	true  = 1
} bool;

#define FAIL    false
#define OFF     false
#define UNSET   false
#define DMA_RD  false
#define OK      true
#define SET     true
#define ON      true
#define DMA_WR  true


typedef struct {
	char name[ NAME_MAX + 1 ];
	int pad;
	int sad;
	int eos;
	int eosmode;
	int timo;
	int flags;
	char ename[ PATH_MAX + 1 ];
	char file[ PATH_MAX + 1 ];
} Device;


typedef struct {
	unsigned long  cnt;        /* I/O count                */
	int            arg;        /* other arguments          */
	void          *buf;        /* I/O buffer               */
	int            stat;       /* returned status          */
	int            err;        /* returned error code      */
} GPIB_Arg;


/* Definition the numbers of all the different ioctl() commands - the
   character 'j' as the 'magic' type seems to be not in use for command
   numbers above 0x3F (see /usr/src/linux/Documentation/ioctl-number.txt).
   All commands need write access to the structure passed to ioctl()
   because status and error states are returned via this structure.
*/

#define GPIB_MAGIC_IOC      'j'

#define GPIB_IOC_READ       _IOWR( GPIB_MAGIC_IOC, 0xC0, GPIB_Arg )
#define GPIB_IOC_WRT        _IOWR( GPIB_MAGIC_IOC, 0xC1, GPIB_Arg )
#define GPIB_IOC_CMD        _IOWR( GPIB_MAGIC_IOC, 0xC2, GPIB_Arg )
#define GPIB_IOC_DVRD       _IOWR( GPIB_MAGIC_IOC, 0xC3, GPIB_Arg )
#define GPIB_IOC_DVWRT      _IOWR( GPIB_MAGIC_IOC, 0xC4, GPIB_Arg )
#define GPIB_IOC_WAIT       _IOWR( GPIB_MAGIC_IOC, 0xC5, GPIB_Arg )
#define GPIB_IOC_ONL        _IOR(  GPIB_MAGIC_IOC, 0xC6, GPIB_Arg )
#define GPIB_IOC_SIC        _IOR(  GPIB_MAGIC_IOC, 0xC7, GPIB_Arg )
#define GPIB_IOC_REN        _IOWR( GPIB_MAGIC_IOC, 0xC8, GPIB_Arg )
#define GPIB_IOC_CAC        _IOR(  GPIB_MAGIC_IOC, 0xCA, GPIB_Arg )
#define GPIB_IOC_GTS        _IOR(  GPIB_MAGIC_IOC, 0xCB, GPIB_Arg )
#define GPIB_IOC_TRG        _IOWR( GPIB_MAGIC_IOC, 0xCC, GPIB_Arg )
#define GPIB_IOC_SDC        _IOWR( GPIB_MAGIC_IOC, 0xCD, GPIB_Arg )
#define GPIB_IOC_DCL        _IOR(  GPIB_MAGIC_IOC, 0xCE, GPIB_Arg )
#define GPIB_IOC_LOCAL      _IOWR( GPIB_MAGIC_IOC, 0xCF, GPIB_Arg )
#define GPIB_IOC_RSP        _IOWR( GPIB_MAGIC_IOC, 0xD0, GPIB_Arg )
#define GPIB_IOC_PPC        _IOWR( GPIB_MAGIC_IOC, 0xD1, GPIB_Arg )
#define GPIB_IOC_RPP        _IOWR( GPIB_MAGIC_IOC, 0xD2, GPIB_Arg )
#define GPIB_IOC_RSV        _IOR(  GPIB_MAGIC_IOC, 0xD3, GPIB_Arg )
#define GPIB_IOC_TIMO       _IOWR( GPIB_MAGIC_IOC, 0xD4, GPIB_Arg )
#define GPIB_IOC_EOS        _IOWR( GPIB_MAGIC_IOC, 0xD5, GPIB_Arg )
#define GPIB_IOC_EOT        _IOWR( GPIB_MAGIC_IOC, 0xD6, GPIB_Arg )
#define GPIB_IOC_DMA        _IOWR( GPIB_MAGIC_IOC, 0xD7, GPIB_Arg )
#define GPIB_IOC_IRQ        _IOWR( GPIB_MAGIC_IOC, 0xD8, GPIB_Arg )
#define GPIB_IOC_LINES      _IOWR( GPIB_MAGIC_IOC, 0xD9, GPIB_Arg )
#define GPIB_IOC_PCT        _IOR(  GPIB_MAGIC_IOC, 0xDA, GPIB_Arg )
#define GPIB_IOC_PAD        _IOWR( GPIB_MAGIC_IOC, 0xDB, GPIB_Arg )
#define GPIB_IOC_SAD        _IOWR( GPIB_MAGIC_IOC, 0xDC, GPIB_Arg )
#define GPIB_IOC_STATE      _IOR(  GPIB_MAGIC_IOC, 0xDD, GPIB_Arg )
#define GPIB_IOC_LN         _IOWR( GPIB_MAGIC_IOC, 0xDE, GPIB_Arg )
#define GPIB_IOC_LLO        _IOWR( GPIB_MAGIC_IOC, 0xDF, GPIB_Arg )


#define NO_SAD       0
#define ALL_SAD     -1

/*
 * Local variables:
 * c-basic-offset: 8
 * c-indent-level: 8
 * c-brace-imaginary-offset: 0
 * c-brace-offset: 0
 * c-argdecl-indent: 4
 * c-label-ofset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: t
 * tab-width: 8
 * End:
 */
