/*
  $Id: util.c 736 2006-07-07 15:48:51Z jens $

  Copyright (C) 2001-2006 Jens Thoms Toerring
 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
  To contact the author send email to: jt@toerring.de
*/


#include "lib.h"


extern Device devices[ ];
extern int num_devices;
extern int master_dev;

extern int gpib_status;
extern int gpib_error;
extern unsigned long gpib_count;

extern bool board_is_online;


int gpib_tmo( int dev_dscr,
              int timo );
int gpib_eos( int dev_dscr,
              int eos_byte );
int gpib_eos_char( int dev_dscr,
                   int eos_char );
int gpib_eos_mode( int dev_dscr,
                   int eos_mode );

int gpib_setup_timo( Device * dev );
int gpib_setup_eos( Device * dev );
Device *check_dscr( int dev_dscr );


extern int gpib_ioctl( int cmd,
                       ... );


/*---------------------------------------------------------------*
 * This function is called always before all functions involving
 * data transfers between the board and a device. It sets up the
 * timeout interval to the value stored in the device structure
 * passed as argument. Several functions that just send commands
 * to devices (as gpib_loc(), gpib_trg() etc.) use the timeout
 * value for the master device - to make it simpler to get at
 * this value the timeout of the master device is automatically
 * set if the function is called with a NULL pointer.
 *---------------------------------------------------------------*/

int gpib_setup_timo( Device * dev )
{
        return gpib_ioctl( GPIB_IOC_TIMO, dev->timo );
}


/*-----------------------------------------------------------------*
 * This function changes the timeout value stored in the structure
 * for a device but doesn't change directly any board settings
 * The old setting is returned in 'gpib_count'.
 *-----------------------------------------------------------------*/

int gpib_tmo( int dev_dscr,
              int timo )
{
        Device *dev;


        if ( ( dev = check_dscr( dev_dscr ) ) == NULL )
                return gpib_status;

        /* Check that the timeout value is reasonable */

        if ( timo < TNONE || timo > T1000s ) {
                gpib_status |= GPIB_ERR;
                gpib_error   = GPIB_EARG;
                return gpib_status;
        }

        /* Update the entry in the device structure */

        gpib_count = dev->timo;
        dev->timo = timo;

        return gpib_status;
}


/*--------------------------------------------------------------------------*
 * Function sets the EOS character and the eosmode in the device structure.
 * This is not send to the board yet but the new values are used on the
 * next read from or write to the device.
 *--------------------------------------------------------------------------*/

int gpib_eos( int dev_dscr,
              int eos_byte )
{
        Device *dev;


        if ( ( dev = check_dscr( dev_dscr ) ) == NULL )
                return gpib_status;

        dev->eos = eos_byte & 0xff;
        dev->eosmode = ( eos_byte >> 8 ) & 0xff;

        return gpib_status;
}


/*------------------------------------------------------*
 * Function sets a new EOS byte for the selected device
 *------------------------------------------------------*/

int gpib_eos_char( int dev_dscr,
                   int eos_char )
{
        Device *dev;

        if ( ( dev = check_dscr( dev_dscr ) ) == NULL )
                return gpib_status;

        dev->eos = eos_char & 0xff;
        return gpib_status;
}


/*------------------------------------------------------*
 * Function sets a new EOS mode for the selected device
 *------------------------------------------------------*/

int gpib_eos_mode( int dev_dscr,
                   int eos_mode )
{
        Device *dev;

        if ( ( dev = check_dscr( dev_dscr ) ) == NULL )
                return gpib_status;

        dev->eosmode = eos_mode & 0xff;
        return gpib_status;
}


/*-----------------------------------------------------------------------*
 * This function is always called before a transfer of data from or to a
 * device. It sets up the board to use the EOS character and eosmode as
 * stored in the device structure for the following transfer.
 *-----------------------------------------------------------------------*/

int gpib_setup_eos( Device * dev )
{
        return gpib_ioctl( GPIB_IOC_EOS, ( ( dev->eosmode & 0xff ) << 8 ) |
                                         ( dev->eos & 0xff ) );
}


/*--------------------------------------------------------------------*
 * Check the device descriptor - if the board hasn't been initialized
 * by a call off gpib_find() this will always return an error.
 *--------------------------------------------------------------------*/

Device * check_dscr( int dev_dscr )
{
        if ( ! board_is_online ) {
                gpib_status &= ~ ( GPIB_CMPL | GPIB_END | GPIB_TIMO );
                gpib_status |= GPIB_ERR;
                gpib_error   = GPIB_EDVR;
                return NULL;
        }

        gpib_ioctl( GPIB_IOC_STATE );

        if ( dev_dscr < 0 || dev_dscr >= num_devices ||
             ! ( devices[ dev_dscr ].flags & IS_INIT ) ) {
                gpib_status |= GPIB_ERR;
                gpib_error   = GPIB_EDVR;
                return NULL;
        }

        return &devices[ dev_dscr ];
}


/*
 * Local variables:
 * c-basic-offset: 8
 * c-indent-level: 8
 * c-brace-imaginary-offset: 0
 * c-brace-offset: 0
 * c-argdecl-indent: 4
 * c-label-ofset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: nil
 * tab-width: 8
 * End:
 */
